# Copyright (C) 2008-2013 One Laptop per Child
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


"""
Activity information microformat parser.

This updater backend pulls activity information from a HTML page that
embeds activity information in special CSS selectors. For more info, see
http://wiki.sugarlabs.org/go/Activity_Team/Activity_Microformat
"""

import os
import locale
import logging
from tempfile import NamedTemporaryFile

from io import StringIO
from configparser import ConfigParser
from zipfile import ZipFile
from urllib.parse import urljoin
from html.parser import HTMLParser

from gi.repository import GLib
from gi.repository import GObject
from gi.repository import Gio

from sugar3.bundle.bundleversion import NormalizedVersion, InvalidVersionError

from jarabe.util import httprange
from jarabe.model import bundleregistry
from jarabe.model.update import BundleUpdate
from jarabe.util.downloader import Downloader

_logger = logging.getLogger('microformat')
_MICROFORMAT_URL_PATH = 'org.sugarlabs.update'
_MICROFORMAT_URL_KEY = 'microformat-update-url'


class _UpdateHTMLParser(HTMLParser):
    """HTML parser to pull out data expressed in our microformat."""

    def __init__(self, base_href):
        HTMLParser.__init__(self)
        self.base_href = base_href
        self.results = {}
        self.group_name = self.group_desc = None
        self.in_group_name = self.in_group_desc = self.in_activity = 0
        self._clear_info()

    def _clear_info(self):
        self.in_activity_id = self.in_activity_url = 0
        self.in_activity_version = 0
        self.in_activity_optional = 0
        self.last_id = self.last_version = self.last_url = None
        self.last_optional = False

    def handle_starttag(self, tag, attrs):
        classes = ' '.join([val for attr, val in attrs if attr == 'class'])\
                  .split()
        if self.in_group_name == 0:
            if ('id', 'olpc-activity-group-name') in attrs:
                self.in_group_name = 1
        else:
            self.in_group_name += 1

        if self.in_group_desc == 0:
            if ('id', 'olpc-activity-group-desc') in attrs:
                self.in_group_desc = 1
        else:
            self.in_group_desc += 1

        if self.in_activity == 0:
            if 'olpc-activity-info' in classes:
                self.in_activity = 1
                self._clear_info()
            return

        self.in_activity += 1
        if self.in_activity_id == 0:
            if 'olpc-activity-id' in classes:
                self.in_activity_id = 1
        else:
            self.in_activity_id += 1

        if self.in_activity_version == 0:
            if 'olpc-activity-version' in classes:
                self.in_activity_version = 1
        else:
            self.in_activity_version += 1

        if self.in_activity_url == 0:
            if 'olpc-activity-url' in classes:
                self.in_activity_url = 1
        else:
            self.in_activity_url += 1

        if self.in_activity_optional == 0:
            if 'olpc-activity-optional' in classes:
                self.in_activity_optional = 1
        else:
            self.in_activity_optional += 1

        # an href inside activity_url is the droid we are looking for.
        if self.in_activity_url > 0:
            for a, v in attrs:
                if a == 'href':
                    self.last_url = urljoin(self.base_href, v)
                    break

    def handle_data(self, data):
        if self.in_group_name:
            self.group_name = data.strip()

        if self.in_group_desc:
            self.group_desc = data.strip()

        if self.in_activity_id > 0:
            if self.last_id is None:
                self.last_id = data.strip()
            else:
                self.last_id = self.last_id + data.strip()

        if self.in_activity_version > 0:
            try:
                self.last_version = NormalizedVersion(data.strip())
            except InvalidVersionError:
                pass

        if self.in_activity_optional > 0:
            # a value 1 means that this activity is optional
            self.last_optional = data.strip() == '1'

    def handle_endtag(self, tag):
        if self.in_group_name > 0:
            self.in_group_name -= 1

        if self.in_group_desc > 0:
            self.in_group_desc -= 1

        if self.in_activity_id > 0:
            self.in_activity_id -= 1

        if self.in_activity_version > 0:
            self.in_activity_version -= 1

        if self.in_activity_optional > 0:
            self.in_activity_optional -= 1

        if self.in_activity_url > 0:
            self.in_activity_url -= 1

        if self.in_activity > 0:
            self.in_activity -= 1

            if self.in_activity == 0:
                self._process_result()

    def _process_result(self):
        if self.last_id is None or self.last_id.strip() == '':
            return

        if self.last_version is not None and self.last_url is not None:
            if self.last_id in self.results:
                if self.last_version < self.results[self.last_id][0]:
                    return  # already found a better version
            self.results[self.last_id] = (self.last_version, self.last_url,
                                          self.last_optional)


class MicroformatUpdater(object):
    """
    Microformat updater backend. The rough code flow here is:
     1. Query update URL and parse results
     2. For each activity update:
       a) If we already have this activity installed, use GIO to asynchronously
          lookup the size of the download.
       b) If we don't have the activity installed, use MetadataLookup
          to lookup activity name and size.
    """

    def __init__(self):
        self._icon_temp_files = []

    def _query(self):
        self.clean()
        settings = Gio.Settings.new(_MICROFORMAT_URL_PATH)
        url = settings.get_string(_MICROFORMAT_URL_KEY)
        _logger.debug("Query %s %r", url, url)
        if url == "":
            self._completion_cb([])
            return

        self._parser = _UpdateHTMLParser(url)
        # wiki.laptop.org have agresive cache, we set max-age=600
        # to be sure the page is no older than 10 minutes
        request_headers = {'Cache-Control': 'max-age=600'}
        downloader = Downloader(url, request_headers=request_headers)
        downloader.connect('got-chunk', self._got_chunk_cb)
        downloader.connect('complete', self._complete_cb)
        downloader.download_chunked()
        self._progress_cb(None, 0)

    def _got_chunk_cb(self, downloader, data):
        self._parser.feed(data.get_data())

    def _complete_cb(self, downloader, result):
        if isinstance(result, Exception):
            _logger.warning("Failed to read update info: %s", result)
            self._error_cb(result)
            return

        self._parser.close()
        _logger.debug("Found %d activities", len(self._parser.results))
        self._filter_results()
        self._check_next_update()

    def _filter_results(self):
        # Remove updates for which we already have an equivalent or newer
        # version installed. Queue the remaining ones to be checked.
        registry = bundleregistry.get_registry()
        self._bundles_to_check = []
        for bundle_id, data in list(self._parser.results.items()):
            # filter optional activities for automatic updates
            if self._auto and data[2] is True:
                logging.debug('filtered optional activity %s', bundle_id)
                continue

            bundle = registry.get_bundle(bundle_id)
            if bundle:
                if data[0] <= NormalizedVersion(bundle.get_activity_version()):
                    continue

            name = bundle.get_name() if bundle else None
            bundle_update = BundleUpdate(bundle_id, name, data[0], data[1], 0,
                                         optional=data[2])
            self._bundles_to_check.append(bundle_update)
        self._total_bundles_to_check = len(self._bundles_to_check)
        _logger.debug("%d results after filter", self._total_bundles_to_check)

    def _check_next_update(self):
        if self._cancelling or len(self._bundles_to_check) == 0:
            self._completion_cb(self._updates)
            return

        total = self._total_bundles_to_check
        current = total - len(self._bundles_to_check)
        progress = current / float(total)

        self._bundle_update = self._bundles_to_check.pop()
        _logger.debug("Check %s", self._bundle_update.bundle_id)

        # There is no need for a special name lookup for an automatic update.
        # The name lookup is only for UI purposes, but we are running in the
        # background.
        if self._bundle_update.name is None and self._auto:
            self._bundle_update.name = self._bundle_update.bundle_id

        if self._bundle_update.name is not None:
            # if we know the name, we just perform an asynchronous size check
            _logger.debug("Performing async size lookup")
            size_check = Downloader(self._bundle_update.link)
            size_check.connect('complete', self._size_lookup_cb)
            size_check.get_size()
            self._progress_cb(self._bundle_update.name, progress)
        else:
            # if we don't know the name, we run a metadata lookup and get
            # the size and name that way
            _logger.debug("Performing metadata lookup")
            namelookup = MetadataLookup(self._bundle_update.link)
            namelookup.connect('complete', self._name_lookup_complete)
            namelookup.run()
            self._progress_cb(self._bundle_update.bundle_id, progress)

    def _size_lookup_cb(self, downloader, result):
        if isinstance(result, Exception):
            _logger.warning("Failed to perform size lookup: %s", result)
        else:
            self._bundle_update.size = result
            self._updates.append(self._bundle_update)

        GLib.idle_add(self._check_next_update)

    def _name_lookup_complete(self, lookup, result, size, icon_file_name):
        _logger.debug("Name lookup result: %r", result)
        if icon_file_name is not None:
            self._icon_temp_files.append(icon_file_name)
            logging.debug('Adding temporary file %s to list', icon_file_name)

        if size is None:
            # if the size lookup failed, assume this update is bad
            self._check_next_update()
            return

        if result is None or isinstance(result, Exception):
            # if we failed to find the name, add the update anyway, using the
            # bundle_id as the best name we have
            self._bundle_update.name = self._bundle_update.bundle_id
        else:
            self._bundle_update.name = result

        self._bundle_update.size = size
        if icon_file_name is not None:
            self._bundle_update.icon_file_name = icon_file_name

        self._updates.append(self._bundle_update)
        self._check_next_update()

    def fetch_update_info(self, installed_bundles, auto, progress_cb,
                          completion_cb, error_cb):
        self._completion_cb = completion_cb
        self._progress_cb = progress_cb
        self._error_cb = error_cb
        self._cancelling = False
        self._updates = []
        self._bundles_to_check = []
        self._total_bundles_to_check = 0
        self._auto = auto
        self._query()

    def cancel(self):
        self._cancelling = True

    def clean(self):
        for filename in self._icon_temp_files:
            logging.debug('Removing temporary file %s', filename)
            try:
                os.unlink(filename)
            except OSError:
                pass
        self._icon_temp_files = []


class MetadataLookup(GObject.GObject):
    """
    Look up the localized activity name and size of a bundle.

    This is useful when no previous version of the activity is installed,
    and there is no local source of the activity's name.
    """
    __gsignals__ = {
        'complete': (GObject.SignalFlags.RUN_FIRST,
                     None, (object, object, object)),
    }

    def __init__(self, url):
        GObject.GObject.__init__(self)
        self._url = url
        self._icon_file_name = None
        self._size = None

    def run(self):
        # Perform the name lookup, catch any exceptions, and report the result.
        try:
            name = self._do_name_lookup()
            self._complete(name)
        except Exception as e:
            self._complete(e)

    def _do_name_lookup(self):
        fd = httprange.open(self._url)
        self._size = fd.size()
        return self._name_from_fd(fd)

    def _name_from_fd(self, fd):
        self._zf = ZipFile(fd)
        self._namelist = self._zf.namelist()
        self._prefix = None
        have_activity_info = False
        have_library_info = False
        for path in self._namelist:
            if path.count('/') != 2:
                continue
            if path.endswith('/activity/activity.info'):
                have_activity_info = True
                self._prefix = path.split('/', 1)[0]
                break
            if path.endswith('/library/library.info'):
                have_library_info = True
                self._prefix = path.split('/', 1)[0]
                break

        if self._prefix is None:
            raise Exception("Couldn't find activity prefix")

        # To find the localized name, we first look for an activity.linfo file
        # in the current locale.
        # We fall back on pulling the non-localized name from activity.info,
        # if there is one.
        # The final fallback is pulling the name from library.info; in the
        # case of a content bundle, that name is expected to be already
        # localized according to the content.
        name = self._locale_data_lookup()
        icon_path = None
        if not name and have_activity_info:
            name = self._activity_info_lookup('name')
        if not name and have_library_info:
            name = self._library_info_lookup('name')

        # get icondata
        if have_activity_info:
            icon_name = self._activity_info_lookup('icon')
        if not name and have_library_info:
            icon_name = self._library_info_lookup('icon')
        icon_path = os.path.join(self._prefix, 'activity',
                                 '%s.svg' % icon_name)
        if icon_path is not None and icon_path in self._namelist:
            icon_data = self._zf.read(icon_path)
            # save the icon to a temporary file
            with NamedTemporaryFile(mode='w', suffix='.svg',
                                    delete=False) as iconfile:
                iconfile.write(icon_data)
                self._icon_file_name = iconfile.name
        return name

    def _locale_data_lookup(self):
        lang = locale.getdefaultlocale()[0]
        if lang is None:
            return None

        for f in ('locale/%s/activity.linfo' % lang,
                  'locale/%s/activity.linfo' % lang[:2]):
            filename = os.path.join(self._prefix, f)
            if filename not in self._namelist:
                continue
            cp = ConfigParser()
            cp.read_file(StringIO(self._zf.read(filename)))
            return cp.get('Activity', 'name')
        return None

    def _activity_info_lookup(self, parameter):
        filename = os.path.join(self._prefix, 'activity', 'activity.info')
        cp = ConfigParser()
        cp.read_file(StringIO(self._zf.read(filename)))
        if cp.has_option('Activity', parameter):
            return cp.get('Activity', parameter)
        return ''

    def _library_info_lookup(self, parameter):
        filename = os.path.join(self._prefix, 'library', 'library.info')
        cp = ConfigParser()
        cp.read_file(StringIO(self._zf.read(filename)))
        if cp.has_option('Library', parameter):
            return cp.get('Library', parameter)
        return ''

    def _complete(self, result):
        GLib.idle_add(self.emit, 'complete', result, self._size,
                      self._icon_file_name)



# Copyright (C) 2013, Gonzalo Odiard
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from gi.repository import GObject
from gi.repository import Gtk

from gettext import gettext as _

from jarabe.journal import model

DS_DBUS_SERVICE = 'org.laptop.sugar.DataStore'
DS_DBUS_INTERFACE = 'org.laptop.sugar.DataStore'
DS_DBUS_PATH = '/org/laptop/sugar/DataStore'


class IconModel(GObject.GObject, Gtk.TreeModel, Gtk.TreeDragSource):
    __gtype_name__ = 'JournalIconModel'

    __gsignals__ = {
        'ready': (GObject.SignalFlags.RUN_FIRST, None, ([])),
        'progress': (GObject.SignalFlags.RUN_FIRST, None, ([])),
    }

    COLUMN_UID = 0
    COLUMN_TITLE = 1
    COLUMN_PREVIEW = 2

    _COLUMN_TYPES = {
        COLUMN_UID: str,
        COLUMN_TITLE: str,
        COLUMN_PREVIEW: str,
    }

    _PAGE_SIZE = 100

    def __init__(self, query):
        GObject.GObject.__init__(self)

        self._last_requested_index = None
        self._cached_row = None
        self._result_set = model.find(query, IconModel._PAGE_SIZE)
        self._temp_drag_file_path = None

        # HACK: The view will tell us that it is resizing so the model can
        # avoid hitting D-Bus and disk.
        self.view_is_resizing = False

        self._result_set.ready.connect(self.__result_set_ready_cb)
        self._result_set.progress.connect(self.__result_set_progress_cb)

    def __result_set_ready_cb(self, **kwargs):
        self.emit('ready')

    def __result_set_progress_cb(self, **kwargs):
        self.emit('progress')

    def setup(self):
        self._result_set.setup()

    def stop(self):
        self._result_set.stop()

    def get_metadata(self, path):
        return model.get(self[path][IconModel.COLUMN_UID])

    def do_get_n_columns(self):
        return len(IconModel._COLUMN_TYPES)

    def do_get_column_type(self, index):
        return IconModel._COLUMN_TYPES[index]

    def do_iter_n_children(self, iterator):
        if iterator is None:
            return self._result_set.length
        return 0

    def do_get_value(self, iterator, column):
        if self.view_is_resizing:
            return None

        index = iterator.user_data
        if index == self._last_requested_index:
            return self._cached_row[column]

        if index >= self._result_set.length:
            return None

        self._result_set.seek(index)
        metadata = self._result_set.read()

        self._last_requested_index = index
        self._cached_row = []
        self._cached_row.append(metadata['uid'])

        title = GObject.markup_escape_text(metadata.get('title',
                                                        _('Untitled')))
        self._cached_row.append(title)

        self._cached_row.append(metadata.get('preview', ''))

        return self._cached_row[column]

    def do_iter_nth_child(self, parent_iter, n):
        return (False, None)

    def do_get_path(self, iterator):
        treepath = Gtk.TreePath((iterator.user_data,))
        return treepath

    def do_get_iter(self, path):
        idx = path.get_indices()[0]
        iterator = Gtk.TreeIter()
        iterator.user_data = idx
        return (True, iterator)

    def do_iter_next(self, iterator):
        idx = iterator.user_data + 1
        if idx >= self._result_set.length:
            iterator.stamp = -1
            return (False, iterator)
        iterator.user_data = idx
        return (True, iterator)

    def do_get_flags(self):
        return Gtk.TreeModelFlags.ITERS_PERSIST | Gtk.TreeModelFlags.LIST_ONLY

    def do_iter_children(self, iterator):
        return (False, iterator)

    def do_iter_has_child(self, iterator):
        return False

    def do_iter_parent(self, iterator):
        return (False, Gtk.TreeIter())



# Copyright (C) 2010 Collabora Ltd. <http://www.collabora.co.uk/>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
from functools import partial
from hashlib import sha1

from gi.repository import GObject
from gi.repository import Gio
import dbus
from dbus import PROPERTIES_IFACE
from gi.repository import TelepathyGLib
ACCOUNT = TelepathyGLib.IFACE_ACCOUNT
ACCOUNT_MANAGER = TelepathyGLib.IFACE_ACCOUNT_MANAGER
CHANNEL = TelepathyGLib.IFACE_CHANNEL
CHANNEL_INTERFACE_GROUP = TelepathyGLib.IFACE_CHANNEL_INTERFACE_GROUP
CHANNEL_TYPE_CONTACT_LIST = TelepathyGLib.IFACE_CHANNEL_TYPE_CONTACT_LIST
CHANNEL_TYPE_FILE_TRANSFER = TelepathyGLib.IFACE_CHANNEL_TYPE_FILE_TRANSFER
CLIENT = TelepathyGLib.IFACE_CLIENT
CONNECTION = TelepathyGLib.IFACE_CONNECTION
CONNECTION_INTERFACE_ALIASING = \
    TelepathyGLib.IFACE_CONNECTION_INTERFACE_ALIASING
CONNECTION_INTERFACE_CONTACTS = \
    TelepathyGLib.IFACE_CONNECTION_INTERFACE_CONTACTS
CONNECTION_INTERFACE_CONTACT_CAPABILITIES = \
    TelepathyGLib.IFACE_CONNECTION_INTERFACE_CONTACT_CAPABILITIES
CONNECTION_INTERFACE_REQUESTS = \
    TelepathyGLib.IFACE_CONNECTION_INTERFACE_REQUESTS
CONNECTION_INTERFACE_SIMPLE_PRESENCE = \
    TelepathyGLib.IFACE_CONNECTION_INTERFACE_SIMPLE_PRESENCE
HANDLE_TYPE_CONTACT = TelepathyGLib.HandleType.CONTACT
HANDLE_TYPE_LIST = TelepathyGLib.HandleType.LIST
CONNECTION_PRESENCE_TYPE_OFFLINE = TelepathyGLib.ConnectionPresenceType.OFFLINE
CONNECTION_STATUS_CONNECTED = TelepathyGLib.ConnectionStatus.CONNECTED
CONNECTION_STATUS_DISCONNECTED = TelepathyGLib.ConnectionStatus.DISCONNECTED

from sugar3.graphics.xocolor import XoColor
from sugar3.profile import get_profile

from jarabe.model.buddy import BuddyModel, get_owner_instance
from jarabe.model import bundleregistry
from jarabe.model import shell


ACCOUNT_MANAGER_SERVICE = 'org.freedesktop.Telepathy.AccountManager'
ACCOUNT_MANAGER_PATH = '/org/freedesktop/Telepathy/AccountManager'
CHANNEL_DISPATCHER_SERVICE = 'org.freedesktop.Telepathy.ChannelDispatcher'
CHANNEL_DISPATCHER_PATH = '/org/freedesktop/Telepathy/ChannelDispatcher'
SUGAR_CLIENT_SERVICE = 'org.freedesktop.Telepathy.Client.Sugar'
SUGAR_CLIENT_PATH = '/org/freedesktop/Telepathy/Client/Sugar'

CONNECTION_INTERFACE_BUDDY_INFO = 'org.laptop.Telepathy.BuddyInfo'
CONNECTION_INTERFACE_ACTIVITY_PROPERTIES = \
    'org.laptop.Telepathy.ActivityProperties'

_QUERY_DBUS_TIMEOUT = 200
"""
Time in seconds to wait when querying contact properties. Some jabber servers
will be very slow in returning these queries, so just be patient.
"""

_model = None


class ActivityModel(GObject.GObject):
    __gsignals__ = {
        'current-buddy-added': (GObject.SignalFlags.RUN_FIRST, None,
                                ([object])),
        'current-buddy-removed': (GObject.SignalFlags.RUN_FIRST, None,
                                  ([object])),
        'buddy-added': (GObject.SignalFlags.RUN_FIRST, None,
                        ([object])),
        'buddy-removed': (GObject.SignalFlags.RUN_FIRST, None,
                          ([object])),
    }

    def __init__(self, activity_id, room_handle):
        GObject.GObject.__init__(self)

        self.activity_id = activity_id
        self.room_handle = room_handle
        self._bundle = None
        self._color = None
        self._private = True
        self._name = None
        self._current_buddies = []
        self._buddies = []

    def get_color(self):
        return self._color

    def set_color(self, color):
        self._color = color

    color = GObject.Property(type=object, getter=get_color, setter=set_color)

    def get_bundle(self):
        return self._bundle

    def set_bundle(self, bundle):
        self._bundle = bundle

    bundle = GObject.Property(type=object, getter=get_bundle,
                              setter=set_bundle)

    def get_name(self):
        return self._name

    def set_name(self, name):
        self._name = name

    name = GObject.Property(type=object, getter=get_name, setter=set_name)

    def is_private(self):
        return self._private

    def set_private(self, private):
        self._private = private

    private = GObject.Property(type=object, getter=is_private,
                               setter=set_private)

    def get_buddies(self):
        return self._buddies

    def add_buddy(self, buddy):
        self._buddies.append(buddy)
        self.notify('buddies')
        self.emit('buddy-added', buddy)

    def remove_buddy(self, buddy):
        self._buddies.remove(buddy)
        self.notify('buddies')
        self.emit('buddy-removed', buddy)

    buddies = GObject.Property(type=object, getter=get_buddies)

    def get_current_buddies(self):
        return self._current_buddies

    def add_current_buddy(self, buddy):
        self._current_buddies.append(buddy)
        self.notify('current-buddies')
        self.emit('current-buddy-added', buddy)

    def remove_current_buddy(self, buddy):
        self._current_buddies.remove(buddy)
        self.notify('current-buddies')
        self.emit('current-buddy-removed', buddy)

    current_buddies = GObject.Property(type=object, getter=get_current_buddies)


class _Account(GObject.GObject):
    __gsignals__ = {
        'activity-added': (GObject.SignalFlags.RUN_FIRST, None,
                           ([object, object])),
        'activity-updated': (GObject.SignalFlags.RUN_FIRST, None,
                             ([object, object])),
        'activity-removed': (GObject.SignalFlags.RUN_FIRST, None,
                             ([object])),
        'buddy-added': (GObject.SignalFlags.RUN_FIRST, None,
                        ([object, object, object])),
        'buddy-updated': (GObject.SignalFlags.RUN_FIRST, None,
                          ([object, object])),
        'buddy-removed': (GObject.SignalFlags.RUN_FIRST, None,
                          ([object])),
        'buddy-joined-activity': (GObject.SignalFlags.RUN_FIRST, None,
                                  ([object, object])),
        'buddy-left-activity': (GObject.SignalFlags.RUN_FIRST, None,
                                ([object, object])),
        'current-activity-updated': (GObject.SignalFlags.RUN_FIRST,
                                     None, ([object, object])),
        'connected': (GObject.SignalFlags.RUN_FIRST, None, ([])),
        'disconnected': (GObject.SignalFlags.RUN_FIRST, None, ([])),
    }

    def __init__(self, account_path):
        GObject.GObject.__init__(self)

        self.object_path = account_path

        self._connection = None
        self._buddy_handles = {}
        self._activity_handles = {}
        self._self_handle = None

        self._buddies_per_activity = {}
        self._activities_per_buddy = {}

        self._home_changed_hid = None

        self._start_listening()

    def _close_connection(self):
        self._connection = None
        if self._home_changed_hid is not None:
            model = shell.get_model()
            model.disconnect(self._home_changed_hid)
            self._home_changed_hid = None

    def _start_listening(self):
        bus = dbus.Bus()
        obj = bus.get_object(ACCOUNT_MANAGER_SERVICE, self.object_path)
        obj.Get(ACCOUNT, 'Connection',
                reply_handler=self.__got_connection_cb,
                error_handler=partial(self.__error_handler_cb,
                                      'Account.GetConnection'))
        obj.connect_to_signal(
            'AccountPropertyChanged', self.__account_property_changed_cb)

    def __error_handler_cb(self, function_name, error):
        raise RuntimeError('Error when calling %s: %s' % (function_name,
                                                          error))

    def __got_connection_cb(self, connection_path):
        logging.debug('_Account.__got_connection_cb %r', connection_path)

        if connection_path == '/':
            self._check_registration_error()
            return

        self._prepare_connection(connection_path)

    def _check_registration_error(self):
        """
        See if a previous connection attempt failed and we need to unset
        the register flag.
        """
        bus = dbus.Bus()
        obj = bus.get_object(ACCOUNT_MANAGER_SERVICE, self.object_path)
        obj.Get(ACCOUNT, 'ConnectionError',
                reply_handler=self.__got_connection_error_cb,
                error_handler=partial(self.__error_handler_cb,
                                      'Account.GetConnectionError'))

    def __got_connection_error_cb(self, error):
        logging.debug('_Account.__got_connection_error_cb %r', error)
        if error == 'org.freedesktop.Telepathy.Error.RegistrationExists':
            bus = dbus.Bus()
            obj = bus.get_object(ACCOUNT_MANAGER_SERVICE, self.object_path)
            obj.UpdateParameters({'register': False}, [],
                                 dbus_interface=ACCOUNT)

    def __account_property_changed_cb(self, properties):
        logging.debug('_Account.__account_property_changed_cb %r %r %r',
                      self.object_path, properties.get('Connection', None),
                      self._connection)
        if 'Connection' not in properties:
            return
        if properties['Connection'] == '/':
            self._check_registration_error()
            self._close_connection()
        elif self._connection is None:
            self._prepare_connection(properties['Connection'])

    def _prepare_connection(self, connection_path):
        connection_name = connection_path.replace('/', '.')[1:]

        self._connection = {}
        self._object_path = connection_path
        self.conn_ready = False
        self.conn_proxy = dbus.Bus().get_object(
            connection_name, connection_path)
        self._connection[PROPERTIES_IFACE] = dbus.Interface(
            self.conn_proxy, PROPERTIES_IFACE)
        self._connection[CONNECTION_INTERFACE_ALIASING] = \
            dbus.Interface(self.conn_proxy, CONNECTION_INTERFACE_ALIASING)
        self._connection[CONNECTION_INTERFACE_SIMPLE_PRESENCE] = \
            dbus.Interface(self.conn_proxy,
                           CONNECTION_INTERFACE_SIMPLE_PRESENCE)
        self._connection[CONNECTION_INTERFACE_REQUESTS] = \
            dbus.Interface(self.conn_proxy, CONNECTION_INTERFACE_REQUESTS)
        self._connection[CONNECTION_INTERFACE_ACTIVITY_PROPERTIES] = \
            dbus.Interface(self.conn_proxy,
                           CONNECTION_INTERFACE_ACTIVITY_PROPERTIES)
        self._connection[CONNECTION_INTERFACE_CONTACTS] = \
            dbus.Interface(self.conn_proxy, CONNECTION_INTERFACE_CONTACTS)
        self._connection[CONNECTION] = dbus.Interface(
            self.conn_proxy, CONNECTION)
        self._connection[CONNECTION].GetInterfaces(
            reply_handler=self.__conn_get_interfaces_reply_cb,
            error_handler=partial(
                self.__error_handler_cb,
                'dbus.GetInterfaces'))

    def __conn_get_interfaces_reply_cb(self, interfaces):
        for interface in interfaces:
            self._connection[interface] = dbus.Interface(
                self.conn_proxy, interface)
        self.conn_ready = True
        self.__connection_ready_cb(self._connection)

    def __connection_ready_cb(self, connection):
        if not self.conn_ready:
            return

        logging.debug('_Account.__connection_ready_cb %r',
                      self._object_path)
        connection[CONNECTION].connect_to_signal('StatusChanged',
                                                 self.__status_changed_cb)

        connection[PROPERTIES_IFACE].Get(CONNECTION,
                                         'Status',
                                         reply_handler=self.__get_status_cb,
                                         error_handler=partial(
                                             self.__error_handler_cb,
                                             'Connection.GetStatus'))

    def __get_status_cb(self, status):
        logging.debug('_Account.__get_status_cb %r %r',
                      self._object_path, status)
        self._update_status(status)

    def __status_changed_cb(self, status, reason):
        logging.debug('_Account.__status_changed_cb %r %r', status, reason)
        self._update_status(status)

    def _update_status(self, status):
        if status == CONNECTION_STATUS_CONNECTED:
            self._connection[PROPERTIES_IFACE].Get(
                CONNECTION,
                'SelfHandle',
                reply_handler=self.__get_self_handle_cb,
                error_handler=partial(
                    self.__error_handler_cb,
                    'Connection.GetSelfHandle'))
            self.emit('connected')
        else:
            for contact_handle, contact_id in list(
                    self._buddy_handles.items()):
                if contact_id is not None:
                    self.emit('buddy-removed', contact_id)

            for room_handle, activity_id in list(
                    self._activity_handles.items()):
                self.emit('activity-removed', activity_id)

            self._buddy_handles = {}
            self._activity_handles = {}
            self._buddies_per_activity = {}
            self._activities_per_buddy = {}

            self.emit('disconnected')

        if status == CONNECTION_STATUS_DISCONNECTED:
            self._close_connection()

    def __get_self_handle_cb(self, self_handle):
        self._self_handle = self_handle

        if CONNECTION_INTERFACE_CONTACT_CAPABILITIES in self._connection:
            interface = CONNECTION_INTERFACE_CONTACT_CAPABILITIES
            connection = self._connection[interface]
            client_name = CLIENT + '.Sugar.FileTransfer'
            file_transfer_channel_class = {
                CHANNEL + '.ChannelType': CHANNEL_TYPE_FILE_TRANSFER,
                CHANNEL + '.TargetHandleType': HANDLE_TYPE_CONTACT}
            capabilities = []
            connection.UpdateCapabilities(
                [(client_name, [file_transfer_channel_class], capabilities)],
                reply_handler=self.__update_capabilities_cb,
                error_handler=partial(self.__error_handler_cb,
                                      'Connection.UpdateCapabilities'))

        connection = self._connection[CONNECTION_INTERFACE_ALIASING]
        connection.connect_to_signal('AliasesChanged',
                                     self.__aliases_changed_cb)

        connection = self._connection[CONNECTION_INTERFACE_SIMPLE_PRESENCE]
        connection.connect_to_signal('PresencesChanged',
                                     self.__presences_changed_cb)

        if CONNECTION_INTERFACE_BUDDY_INFO in self._connection:
            connection = self._connection[CONNECTION_INTERFACE_BUDDY_INFO]
            connection.connect_to_signal('PropertiesChanged',
                                         self.__buddy_info_updated_cb,
                                         byte_arrays=True)

            connection.connect_to_signal('ActivitiesChanged',
                                         self.__buddy_activities_changed_cb)

            connection.connect_to_signal('CurrentActivityChanged',
                                         self.__current_activity_changed_cb)

            if self._home_changed_hid is None:
                home_model = shell.get_model()
                self._home_changed_hid = home_model.connect(
                    'active-activity-changed',
                    self.__active_activity_changed_cb)
        else:
            logging.warning('Connection %s does not support OLPC buddy '
                            'properties', self._object_path)

        if CONNECTION_INTERFACE_ACTIVITY_PROPERTIES in self._connection:
            connection = self._connection[
                CONNECTION_INTERFACE_ACTIVITY_PROPERTIES]
            connection.connect_to_signal(
                'ActivityPropertiesChanged',
                self.__activity_properties_changed_cb)
        else:
            logging.warning('Connection %s does not support OLPC activity '
                            'properties', self._object_path)

        properties = {
            CHANNEL + '.ChannelType': CHANNEL_TYPE_CONTACT_LIST,
            CHANNEL + '.TargetHandleType': HANDLE_TYPE_LIST,
            CHANNEL + '.TargetID': 'subscribe',
        }
        properties = dbus.Dictionary(properties, signature='sv')
        connection = self._connection[CONNECTION_INTERFACE_REQUESTS]
        is_ours, channel_path, properties = \
            connection.EnsureChannel(properties)

        channel = {}
        service_name = self._object_path.replace('/', '.')[1:]
        text_proxy = dbus.Bus().get_object(service_name, channel_path)
        channel[PROPERTIES_IFACE] = dbus.Interface(
            text_proxy, PROPERTIES_IFACE)
        channel[CHANNEL_INTERFACE_GROUP] = \
            dbus.Interface(text_proxy, CHANNEL_INTERFACE_GROUP)

        channel[CHANNEL_INTERFACE_GROUP].connect_to_signal(
            'MembersChanged', self.__members_changed_cb)

        channel[PROPERTIES_IFACE].Get(
            CHANNEL_INTERFACE_GROUP,
            'Members',
            reply_handler=self.__get_members_ready_cb,
            error_handler=partial(
                self.__error_handler_cb,
                'Connection.GetMembers'))

    def __active_activity_changed_cb(self, model, home_activity):
        if home_activity is None:
            return

        room_handle = 0
        home_activity_id = home_activity.get_activity_id()
        for handle, activity_id in list(self._activity_handles.items()):
            if home_activity_id == activity_id:
                room_handle = handle
                break
        if room_handle == 0:
            home_activity_id = ''

        connection = self._connection[CONNECTION_INTERFACE_BUDDY_INFO]
        connection.SetCurrentActivity(
            home_activity_id,
            room_handle,
            reply_handler=self.__set_current_activity_cb,
            error_handler=self.__set_current_activity_error_cb)

    def __set_current_activity_cb(self):
        logging.debug('_Account.__set_current_activity_cb')

    def __set_current_activity_error_cb(self, error):
        logging.debug('_Account.__set_current_activity__error_cb %r', error)

    def __update_capabilities_cb(self):
        pass

    def __aliases_changed_cb(self, aliases):
        logging.debug('_Account.__aliases_changed_cb')
        for handle, alias in aliases:
            if handle in self._buddy_handles:
                logging.debug('Got handle %r with nick %r, going to update',
                              handle, alias)
                properties = {CONNECTION_INTERFACE_ALIASING + '/alias': alias}
                self.emit('buddy-updated', self._buddy_handles[handle],
                          properties)

    def __presences_changed_cb(self, presences):
        logging.debug('_Account.__presences_changed_cb %r', presences)
        for handle, presence in list(presences.items()):
            if handle in self._buddy_handles:
                presence_type, status_, message_ = presence
                if presence_type == CONNECTION_PRESENCE_TYPE_OFFLINE:
                    contact_id = self._buddy_handles[handle]
                    del self._buddy_handles[handle]
                    self.emit('buddy-removed', contact_id)

    def __buddy_info_updated_cb(self, handle, properties):
        logging.debug('_Account.__buddy_info_updated_cb %r', handle)
        if handle in self._buddy_handles:
            self.emit('buddy-updated', self._buddy_handles[handle], properties)

    def __current_activity_changed_cb(self, contact_handle, activity_id,
                                      room_handle):
        logging.debug('_Account.__current_activity_changed_cb %r %r %r',
                      contact_handle, activity_id, room_handle)
        if contact_handle in self._buddy_handles:
            contact_id = self._buddy_handles[contact_handle]
            if not activity_id and room_handle:
                activity_id = self._activity_handles.get(room_handle, '')
            self.emit('current-activity-updated', contact_id, activity_id)

    def __get_current_activity_cb(self, contact_handle, activity_id,
                                  room_handle):
        logging.debug('_Account.__get_current_activity_cb %r %r %r',
                      contact_handle, activity_id, room_handle)

        if contact_handle in self._buddy_handles:
            contact_id = self._buddy_handles[contact_handle]
            if not activity_id and room_handle:
                activity_id = self._activity_handles.get(room_handle, '')
            self.emit('current-activity-updated', contact_id, activity_id)

    def __buddy_activities_changed_cb(self, buddy_handle, activities):
        self._update_buddy_activities(buddy_handle, activities)

    def _update_buddy_activities(self, buddy_handle, activities):
        logging.debug('_Account._update_buddy_activities')

        if buddy_handle not in self._activities_per_buddy:
            self._activities_per_buddy[buddy_handle] = set()

        for activity_id, room_handle in activities:
            if room_handle not in self._activity_handles:
                self._activity_handles[room_handle] = activity_id

                if buddy_handle == self._self_handle:
                    home_model = shell.get_model()
                    activity = home_model.get_active_activity()
                    if activity.get_activity_id() == activity_id:
                        connection = self._connection[
                            CONNECTION_INTERFACE_BUDDY_INFO]
                        connection.SetCurrentActivity(
                            activity_id,
                            room_handle,
                            reply_handler=self.__set_current_activity_cb,
                            error_handler=self.__set_current_activity_error_cb)

                self.emit('activity-added', room_handle, activity_id)

                connection = self._connection[
                    CONNECTION_INTERFACE_ACTIVITY_PROPERTIES]
                connection.GetProperties(
                    room_handle,
                    reply_handler=partial(self.__get_properties_cb,
                                          room_handle),
                    error_handler=partial(self.__error_handler_cb,
                                          'ActivityProperties.GetProperties'))

                if buddy_handle != self._self_handle:
                    # Sometimes we'll get CurrentActivityChanged before we get
                    # to know about the activity so we miss the event. In that
                    # case, request again the current activity for this buddy.
                    connection = self._connection[
                        CONNECTION_INTERFACE_BUDDY_INFO]
                    connection.GetCurrentActivity(
                        buddy_handle,
                        reply_handler=partial(self.__get_current_activity_cb,
                                              buddy_handle),
                        error_handler=partial(self.__error_handler_cb,
                                              'BuddyInfo.GetCurrentActivity'))

            if activity_id not in self._buddies_per_activity:
                self._buddies_per_activity[activity_id] = set()
            self._buddies_per_activity[activity_id].add(buddy_handle)
            if activity_id not in self._activities_per_buddy[buddy_handle]:
                self._activities_per_buddy[buddy_handle].add(activity_id)
                if buddy_handle != self._self_handle:
                    self.emit('buddy-joined-activity',
                              self._buddy_handles[buddy_handle],
                              activity_id)

        current_activity_ids = \
            [activity_id for activity_id, room_handle in activities]
        for activity_id in self._activities_per_buddy[buddy_handle].copy():
            if activity_id not in current_activity_ids:
                self._remove_buddy_from_activity(buddy_handle, activity_id)

    def __get_properties_cb(self, room_handle, properties):
        logging.debug('_Account.__get_properties_cb %r %r', room_handle,
                      properties)
        if properties:
            self._update_activity(room_handle, properties)

    def _remove_buddy_from_activity(self, buddy_handle, activity_id):
        if buddy_handle in self._buddies_per_activity[activity_id]:
            self._buddies_per_activity[activity_id].remove(buddy_handle)

        if activity_id in self._activities_per_buddy[buddy_handle]:
            self._activities_per_buddy[buddy_handle].remove(activity_id)

        if buddy_handle != self._self_handle:
            self.emit('buddy-left-activity',
                      self._buddy_handles[buddy_handle],
                      activity_id)

        if not self._buddies_per_activity[activity_id]:
            del self._buddies_per_activity[activity_id]

            for room_handle in self._activity_handles.copy():
                if self._activity_handles[room_handle] == activity_id:
                    del self._activity_handles[room_handle]
                    break

            self.emit('activity-removed', activity_id)

    def __activity_properties_changed_cb(self, room_handle, properties):
        logging.debug('_Account.__activity_properties_changed_cb %r %r',
                      room_handle, properties)
        self._update_activity(room_handle, properties)

    def _update_activity(self, room_handle, properties):
        if room_handle in self._activity_handles:
            self.emit('activity-updated', self._activity_handles[room_handle],
                      properties)
        else:
            logging.debug('_Account.__activity_properties_changed_cb unknown '
                          'activity')
            # We don't get ActivitiesChanged for the owner of the connection,
            # so we query for its activities in order to find out.
            if CONNECTION_INTERFACE_BUDDY_INFO in self._connection:
                handle = self._self_handle
                connection = self._connection[CONNECTION_INTERFACE_BUDDY_INFO]
                connection.GetActivities(
                    handle,
                    reply_handler=partial(self.__got_activities_cb, handle),
                    error_handler=partial(self.__error_handler_cb,
                                          'BuddyInfo.Getactivities'))

    def __members_changed_cb(self, message, added, removed, local_pending,
                             remote_pending, actor, reason):
        self._add_buddy_handles(added)

    def __get_members_ready_cb(self, handles):
        logging.debug('_Account.__get_members_ready_cb %r', handles)
        if not handles:
            return

        self._add_buddy_handles(handles)

    def _add_buddy_handles(self, handles):
        logging.debug('_Account._add_buddy_handles %r', handles)
        interfaces = [CONNECTION, CONNECTION_INTERFACE_ALIASING]
        self._connection[CONNECTION_INTERFACE_CONTACTS].GetContactAttributes(
            handles, interfaces, False,
            reply_handler=self.__get_contact_attributes_cb,
            error_handler=partial(self.__error_handler_cb,
                                  'Contacts.GetContactAttributes'))

    def __got_buddy_info_cb(self, handle, nick, properties):
        logging.debug('_Account.__got_buddy_info_cb %r', handle)
        self.emit('buddy-updated', self._buddy_handles[handle], properties)

    def __get_contact_attributes_cb(self, attributes):
        logging.debug('_Account.__get_contact_attributes_cb %r',
                      list(attributes.keys()))

        for handle in list(attributes.keys()):
            nick = attributes[handle][CONNECTION_INTERFACE_ALIASING + '/alias']

            if handle == self._self_handle:
                logging.debug('_Account.__get_contact_attributes_cb,'
                              ' do not add ourself %r', handle)
                continue

            if handle in self._buddy_handles and \
                    not self._buddy_handles[handle] is None:
                logging.debug('Got handle %r with nick %r, going to update',
                              handle, nick)
                self.emit('buddy-updated', self._buddy_handles[handle],
                          attributes[handle])
            else:
                logging.debug('Got handle %r with nick %r, going to add',
                              handle, nick)

                contact_id = attributes[handle][CONNECTION + '/contact-id']
                self._buddy_handles[handle] = contact_id

                if CONNECTION_INTERFACE_BUDDY_INFO in self._connection:
                    connection = \
                        self._connection[CONNECTION_INTERFACE_BUDDY_INFO]

                    connection.GetProperties(
                        handle,
                        reply_handler=partial(self.__got_buddy_info_cb, handle,
                                              nick),
                        error_handler=partial(self.__error_handler_cb,
                                              'BuddyInfo.GetProperties'),
                        byte_arrays=True,
                        timeout=_QUERY_DBUS_TIMEOUT)

                    connection.GetActivities(
                        handle,
                        reply_handler=partial(self.__got_activities_cb,
                                              handle),
                        error_handler=partial(self.__error_handler_cb,
                                              'BuddyInfo.GetActivities'),
                        timeout=_QUERY_DBUS_TIMEOUT)

                    connection.GetCurrentActivity(
                        handle,
                        reply_handler=partial(self.__get_current_activity_cb,
                                              handle),
                        error_handler=partial(self.__error_handler_cb,
                                              'BuddyInfo.GetCurrentActivity'),
                        timeout=_QUERY_DBUS_TIMEOUT)

                self.emit('buddy-added', contact_id, nick, handle)

    def __got_activities_cb(self, buddy_handle, activities):
        logging.debug('_Account.__got_activities_cb %r %r', buddy_handle,
                      activities)
        self._update_buddy_activities(buddy_handle, activities)

    def enable(self):
        logging.debug('_Account.enable %s', self.object_path)
        self._set_enabled(True)

    def disable(self):
        logging.debug('_Account.disable %s', self.object_path)
        self._set_enabled(False)
        self._close_connection()

    def _set_enabled(self, value):
        bus = dbus.Bus()
        obj = bus.get_object(ACCOUNT_MANAGER_SERVICE, self.object_path)
        obj.Set(ACCOUNT, 'Enabled', value,
                reply_handler=self.__set_enabled_cb,
                error_handler=partial(self.__error_handler_cb,
                                      'Account.SetEnabled'),
                dbus_interface=dbus.PROPERTIES_IFACE)

    def __set_enabled_cb(self):
        logging.debug('_Account.__set_enabled_cb success')


class Neighborhood(GObject.GObject):
    __gsignals__ = {
        'activity-added': (GObject.SignalFlags.RUN_FIRST, None,
                           ([object])),
        'activity-removed': (GObject.SignalFlags.RUN_FIRST, None,
                             ([object])),
        'buddy-added': (GObject.SignalFlags.RUN_FIRST, None,
                        ([object])),
        'buddy-removed': (GObject.SignalFlags.RUN_FIRST, None,
                          ([object])),
    }

    def __init__(self):
        GObject.GObject.__init__(self)

        self._buddies = {None: get_owner_instance()}
        self._activities = {}
        self._link_local_account = None
        self._server_account = None
        self._shell_model = shell.get_model()

        self._settings_collaboration = \
            Gio.Settings.new('org.sugarlabs.collaboration')
        self._settings_collaboration.connect(
            'changed::jabber-server', self.__jabber_server_changed_cb)
        self._settings_user = Gio.Settings.new('org.sugarlabs.user')
        self._settings_user.connect(
            'changed::nick', self.__nick_changed_cb)

        bus = dbus.Bus()
        obj = bus.get_object(ACCOUNT_MANAGER_SERVICE, ACCOUNT_MANAGER_PATH)
        account_manager = dbus.Interface(obj, ACCOUNT_MANAGER)
        account_manager.Get(ACCOUNT_MANAGER, 'ValidAccounts',
                            dbus_interface=PROPERTIES_IFACE,
                            reply_handler=self.__got_accounts_cb,
                            error_handler=self.__error_handler_cb)

    def __got_accounts_cb(self, account_paths):
        self._link_local_account = \
            self._ensure_link_local_account(account_paths)
        self._connect_to_account(self._link_local_account)

        self._server_account = self._ensure_server_account(account_paths)
        self._connect_to_account(self._server_account)

    def __error_handler_cb(self, error):
        raise RuntimeError(error)

    def _connect_to_account(self, account):
        account.connect('buddy-added', self.__buddy_added_cb)
        account.connect('buddy-updated', self.__buddy_updated_cb)
        account.connect('buddy-removed', self.__buddy_removed_cb)
        account.connect('buddy-joined-activity',
                        self.__buddy_joined_activity_cb)
        account.connect('buddy-left-activity', self.__buddy_left_activity_cb)
        account.connect('activity-added', self.__activity_added_cb)
        account.connect('activity-updated', self.__activity_updated_cb)
        account.connect('activity-removed', self.__activity_removed_cb)
        account.connect('current-activity-updated',
                        self.__current_activity_updated_cb)
        account.connect('connected', self.__account_connected_cb)
        account.connect('disconnected', self.__account_disconnected_cb)

    def __account_connected_cb(self, account):
        logging.debug('__account_connected_cb %s', account.object_path)
        if account == self._server_account:
            self._link_local_account.disable()

    def __account_disconnected_cb(self, account):
        logging.debug('__account_disconnected_cb %s', account.object_path)
        if account == self._server_account:
            self._link_local_account.enable()

    def _get_published_name(self):
        """Construct the published name based on the public key

        Limit the name to be only 8 characters maximum. The avahi
        service name has a 64 character limit. It consists of
        the room name, the published name and the host name.

        """
        public_key_hash = sha1(
            get_profile().pubkey.encode('utf-8')).hexdigest()
        return public_key_hash[:8]

    def _ensure_link_local_account(self, account_paths):
        for account_path in account_paths:
            if 'salut' in account_path:
                logging.debug('Already have a Salut account')
                account = _Account(account_path)
                account.enable()
                return account

        logging.debug('Still dont have a Salut account, creating one')

        nick = self._settings_user.get_string('nick')

        params = {
            'nickname': nick,
            'first-name': '',
            'last-name': '',
            'jid': self._get_jabber_account_id(),
            'published-name': self._get_published_name(),
        }

        properties = {
            ACCOUNT + '.Enabled': True,
            ACCOUNT + '.Nickname': nick,
            ACCOUNT + '.ConnectAutomatically': True,
        }

        bus = dbus.Bus()
        obj = bus.get_object(ACCOUNT_MANAGER_SERVICE, ACCOUNT_MANAGER_PATH)
        account_manager = dbus.Interface(obj, ACCOUNT_MANAGER)
        account_path = account_manager.CreateAccount('salut', 'local-xmpp',
                                                     'salut', params,
                                                     properties)
        return _Account(account_path)

    def _ensure_server_account(self, account_paths):
        for account_path in account_paths:
            if 'gabble' in account_path:
                logging.debug('Already have a Gabble account')
                account = _Account(account_path)
                account.enable()
                return account

        logging.debug('Still dont have a Gabble account, creating one')

        nick = self._settings_user.get_string('nick')
        server = self._settings_collaboration.get_string('jabber-server')
        key_hash = get_profile().privkey_hash

        params = {
            'account': self._get_jabber_account_id(),
            'password': key_hash,
            'server': server,
            'resource': 'sugar',
            'require-encryption': True,
            'ignore-ssl-errors': True,
            'register': True,
            'old-ssl': True,
            'port': dbus.UInt32(5223),
        }

        properties = {
            ACCOUNT + '.Enabled': True,
            ACCOUNT + '.Nickname': nick,
            ACCOUNT + '.ConnectAutomatically': True,
        }

        bus = dbus.Bus()
        obj = bus.get_object(ACCOUNT_MANAGER_SERVICE, ACCOUNT_MANAGER_PATH)
        account_manager = dbus.Interface(obj, ACCOUNT_MANAGER)
        account_path = account_manager.CreateAccount('gabble', 'jabber',
                                                     'jabber', params,
                                                     properties)
        return _Account(account_path)

    def _get_jabber_account_id(self):
        public_key_hash = sha1(
            get_profile().pubkey.encode('utf-8')).hexdigest()
        server = self._settings_collaboration.get_string('jabber-server')
        return '%s@%s' % (public_key_hash, server)

    def __jabber_server_changed_cb(self, settings, key):
        logging.debug('__jabber_server_changed_cb')

        bus = dbus.Bus()
        account = bus.get_object(ACCOUNT_MANAGER_SERVICE,
                                 self._server_account.object_path)

        server = settings.get_string('jabber-server')
        account_id = self._get_jabber_account_id()
        params_needing_reconnect = account.UpdateParameters(
            {'server': server,
             'account': account_id,
             'register': True},
            dbus.Array([], 's'), dbus_interface=ACCOUNT)
        if params_needing_reconnect:
            account.Reconnect()

        self._update_jid()

    def __nick_changed_cb(self, settings, key):
        logging.debug('__nick_changed_cb')

        nick = settings.get_string('nick')

        bus = dbus.Bus()
        server_obj = bus.get_object(ACCOUNT_MANAGER_SERVICE,
                                    self._server_account.object_path)
        server_obj.Set(ACCOUNT, 'Nickname', nick,
                       dbus_interface=PROPERTIES_IFACE)

        link_local_obj = bus.get_object(ACCOUNT_MANAGER_SERVICE,
                                        self._link_local_account.object_path)
        link_local_obj.Set(ACCOUNT, 'Nickname', nick,
                           dbus_interface=PROPERTIES_IFACE)
        params_needing_reconnect = link_local_obj.UpdateParameters(
            {'nickname': nick, 'published-name': self._get_published_name()},
            dbus.Array([], 's'), dbus_interface=ACCOUNT)
        if params_needing_reconnect:
            link_local_obj.Reconnect()

        self._update_jid()

    def _update_jid(self):
        bus = dbus.Bus()
        account = bus.get_object(ACCOUNT_MANAGER_SERVICE,
                                 self._link_local_account.object_path)

        account_id = self._get_jabber_account_id()
        params_needing_reconnect = account.UpdateParameters(
            {'jid': account_id}, dbus.Array([], 's'), dbus_interface=ACCOUNT)
        if params_needing_reconnect:
            account.Reconnect()

    def __buddy_added_cb(self, account, contact_id, nick, handle):
        logging.debug('__buddy_added_cb %r', contact_id)

        if contact_id in self._buddies:
            logging.debug('__buddy_added_cb buddy already tracked')
            return

        buddy = BuddyModel(
            nick=nick,
            account=account.object_path,
            contact_id=contact_id,
            handle=handle)
        self._buddies[contact_id] = buddy

    def __buddy_updated_cb(self, account, contact_id, properties):
        logging.debug('__buddy_updated_cb %r', contact_id)
        if contact_id is None:
            # Don't know the contact-id yet, will get the full state later
            return

        if contact_id not in self._buddies:
            logging.debug('__buddy_updated_cb Unknown buddy with contact_id'
                          ' %r', contact_id)
            return

        buddy = self._buddies[contact_id]

        is_new = buddy.props.key is None and 'key' in properties

        if 'color' in properties:
            # arrives unicode but we connect with byte_arrays=True - SL #4157
            buddy.props.color = XoColor(str(properties['color']))

        if 'key' in properties:
            buddy.props.key = properties['key']

        nick_key = CONNECTION_INTERFACE_ALIASING + '/alias'
        if nick_key in properties:
            buddy.props.nick = properties[nick_key]

        if is_new:
            self.emit('buddy-added', buddy)

    def __buddy_removed_cb(self, account, contact_id):
        logging.debug('Neighborhood.__buddy_removed_cb %r', contact_id)
        if contact_id not in self._buddies:
            logging.debug('Neighborhood.__buddy_removed_cb Unknown buddy with '
                          'contact_id %r', contact_id)
            return

        buddy = self._buddies[contact_id]
        del self._buddies[contact_id]

        if buddy.props.key is not None:
            self.emit('buddy-removed', buddy)

    def __activity_added_cb(self, account, room_handle, activity_id):
        logging.debug('__activity_added_cb %r %r', room_handle, activity_id)
        if activity_id in self._activities:
            logging.debug('__activity_added_cb activity already tracked')
            return

        activity = ActivityModel(activity_id, room_handle)
        self._activities[activity_id] = activity

    def __activity_updated_cb(self, account, activity_id, properties):
        logging.debug('__activity_updated_cb %r %r', activity_id, properties)
        if activity_id not in self._activities:
            logging.debug('__activity_updated_cb Unknown activity with '
                          'activity_id %r', activity_id)
            return

        registry = bundleregistry.get_registry()
        bundle = registry.get_bundle(properties['type'])
        if not bundle:
            logging.warning('Ignoring shared activity we don''t have')
            return

        activity = self._activities[activity_id]

        is_new = activity.props.bundle is None

        # arrives unicode but we connect with byte_arrays=True - SL #4157
        activity.props.color = XoColor(str(properties['color']))
        activity.props.bundle = bundle
        activity.props.name = properties['name']
        activity.props.private = properties['private']

        if is_new:
            self._shell_model.add_shared_activity(activity_id,
                                                  activity.props.color)
            self.emit('activity-added', activity)

    def __activity_removed_cb(self, account, activity_id):
        logging.debug('__activity_removed_cb %r', activity_id)
        if activity_id not in self._activities:
            logging.debug('Unknown activity with id %s. Already removed?',
                          activity_id)
            return
        activity = self._activities[activity_id]
        del self._activities[activity_id]
        self._shell_model.remove_shared_activity(activity_id)

        if activity.props.bundle is not None:
            self.emit('activity-removed', activity)

    def __current_activity_updated_cb(self, account, contact_id, activity_id):
        logging.debug('__current_activity_updated_cb %r %r', contact_id,
                      activity_id)
        if contact_id not in self._buddies:
            logging.debug('__current_activity_updated_cb Unknown buddy with '
                          'contact_id %r', contact_id)
            return
        if activity_id and activity_id not in self._activities:
            logging.debug('__current_activity_updated_cb Unknown activity with'
                          ' id %s', activity_id)
            activity_id = ''

        buddy = self._buddies[contact_id]
        if buddy.props.current_activity is not None:
            if buddy.props.current_activity.activity_id == activity_id:
                return
            buddy.props.current_activity.remove_current_buddy(buddy)

        if activity_id:
            activity = self._activities[activity_id]
            buddy.props.current_activity = activity
            activity.add_current_buddy(buddy)
        else:
            buddy.props.current_activity = None

    def __buddy_joined_activity_cb(self, account, contact_id, activity_id):
        if contact_id not in self._buddies:
            logging.debug('__buddy_joined_activity_cb Unknown buddy with '
                          'contact_id %r', contact_id)
            return

        if activity_id not in self._activities:
            logging.debug('__buddy_joined_activity_cb Unknown activity with '
                          'activity_id %r', activity_id)
            return

        self._activities[activity_id].add_buddy(self._buddies[contact_id])

    def __buddy_left_activity_cb(self, account, contact_id, activity_id):
        if contact_id not in self._buddies:
            logging.debug('__buddy_left_activity_cb Unknown buddy with '
                          'contact_id %r', contact_id)
            return

        if activity_id not in self._activities:
            logging.debug('__buddy_left_activity_cb Unknown activity with '
                          'activity_id %r', activity_id)
            return

        self._activities[activity_id].remove_buddy(self._buddies[contact_id])

    def get_buddies(self):
        return list(self._buddies.values())

    def get_buddy_by_key(self, key):
        for buddy in list(self._buddies.values()):
            if buddy.key == key:
                return buddy
        return None

    def get_buddy_by_handle(self, contact_handle):
        for buddy in list(self._buddies.values()):
            if not buddy.is_owner() and buddy.handle == contact_handle:
                return buddy
        return None

    def get_activity(self, activity_id):
        return self._activities.get(activity_id, None)

    def get_activity_by_room(self, room_handle):
        for activity in list(self._activities.values()):
            if activity.room_handle == room_handle:
                return activity
        return None

    def get_activities(self):
        return list(self._activities.values())


def get_model():
    global _model
    if _model is None:
        _model = Neighborhood()
    return _model



# Copyright (C) 2011 Walter Bender
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import glob
import hashlib

from gi.repository import Gtk

from sugar3 import profile
from sugar3.activity import bundlebuilder
from sugar3.datastore import datastore
from sugar3.env import get_user_activities_path

import logging
_logger = logging.getLogger('CustomizeBundle')


BADGE_SUBPATH = 'emblems/emblem-view-source.svg'
BADGE_TRANSFORM = '  <g transform="matrix(0.45,0,0,0.45,32,32)">\n'
ICON_TRANSFORM = ' <g transform="matrix(1.0,0,0,1.0,0,0)">\n'
XML_HEADER = '<?xml version="1.0" ?> \
<!DOCTYPE svg  PUBLIC "-//W3C//DTD SVG 1.1//EN" \
"http://www.w3.org/Graphics/SVG/1.1/DTD/svg11.dtd" [\n\
<!ENTITY stroke_color "#010101">\n\
<!ENTITY fill_color "#FFFFFF">\n]>\n'
SVG_START = '<svg enable-background="new 0 0 55 55" height="55px" \
version="1.1" viewBox="0 0 55 55" width="55px" x="0px" xml:space="preserve" \
xmlns="http://www.w3.org/2000/svg" \
xmlns:xlink="http://www.w3.org/1999/xlink" y="0px">\n'
SVG_END = '</svg>\n'


def generate_unique_id():
    """Generate an id based on the user's nick name and their public key
    (Based on schema used by IRC activity).

    """
    nick = profile.get_nick_name()
    pubkey = profile.get_pubkey()
    m = hashlib.sha1()
    m.update(pubkey)
    hexhash = m.hexdigest()

    nick_letters = "".join([x for x in nick if x.isalpha()])

    if not nick_letters:
        nick_letters = 'XO'

    return nick_letters + '_' + hexhash[:4]


def generate_bundle(nick, new_basename):
    """Generate a new .xo bundle for the activity and copy it into the
    Journal.

    """
    new_activity_name = _customize_activity_info(
        nick, new_basename)

    user_activities_path = get_user_activities_path()
    if os.path.exists(os.path.join(user_activities_path, new_basename,
                                   'dist')):
        for path in glob.glob(os.path.join(user_activities_path, new_basename,
                                           'dist', '*')):
            os.remove(path)

    source_dir = os.path.join(user_activities_path, new_basename)
    config = bundlebuilder.Config(
        source_dir=source_dir,
        dist_dir=os.path.join(source_dir, 'dist'),
        dist_name='%s-1' % (new_activity_name))
    bundlebuilder.cmd_dist_xo(config, None)

    dsobject = datastore.create()
    dsobject.metadata['title'] = '%s-1.xo' % (new_activity_name)
    dsobject.metadata['mime_type'] = 'application/vnd.olpc-sugar'
    dsobject.set_file_path(os.path.join(
        user_activities_path, new_basename, 'dist',
        '%s-1.xo' % (new_activity_name)))
    datastore.write(dsobject)
    dsobject.destroy()


def _customize_activity_info(nick, new_basename):
    """Modify bundle_id in new activity.info file:
    (1) change the bundle_id to bundle_id_[NICKNAME];
    (2) change the activity_icon [NICKNAME]-activity-icon.svg;
    (3) set activity_version to 1;
    (4) modify the activity icon by applying a customize overlay.

    """
    new_activity_name = ''
    user_activities_path = get_user_activities_path()

    info_old = open(os.path.join(user_activities_path, new_basename,
                                 'activity', 'activity.info'), 'r')
    info_new = open(os.path.join(user_activities_path, new_basename,
                                 'activity', 'new_activity.info'), 'w')

    for line in info_old:
        if line.find('=') < 0:
            info_new.write(line)
            continue
        name, value = [token.strip() for token in line.split('=', 1)]
        if name == 'bundle_id':
            new_value = '%s_%s' % (value, nick)
        elif name == 'activity_version':
            new_value = '1'
        elif name == 'icon':
            new_value = value
            icon_name = value
        elif name == 'name':
            new_value = '%s_copy_of_%s' % (nick, value)
            new_activity_name = new_value
        else:
            info_new.write(line)
            continue

        info_new.write('%s = %s\n' % (name, new_value))

    info_old.close()
    info_new.close()

    os.rename(os.path.join(user_activities_path, new_basename,
                           'activity', 'new_activity.info'),
              os.path.join(user_activities_path, new_basename,
                           'activity', 'activity.info'))

    _create_custom_icon(new_basename, icon_name)

    return new_activity_name


def _create_custom_icon(new_basename, icon_name):
    """Modify activity icon by overlaying a badge:
    (1) Extract the payload from the badge icon;
    (2) Add a transform to resize it and position it;
    (3) Insert it into the activity icon.

    """
    user_activities_path = get_user_activities_path()
    badge_path = None
    for path in Gtk.IconTheme.get_default().get_search_path():
        if os.path.exists(os.path.join(path, 'sugar', 'scalable',
                                       BADGE_SUBPATH)):
            badge_path = path
            break

    if badge_path is None:
        _logger.debug('%s not found', BADGE_SUBPATH)
        return

    badge_fd = open(os.path.join(badge_path, 'sugar', 'scalable',
                                 BADGE_SUBPATH), 'r')
    badge_payload = _extract_svg_payload(badge_fd)
    badge_fd.close()

    badge_svg = BADGE_TRANSFORM + badge_payload + '\n</g>'

    icon_path = os.path.join(user_activities_path, new_basename, 'activity',
                             icon_name + '.svg')
    icon_fd = open(icon_path, 'r')
    icon_payload = _extract_svg_payload(icon_fd)
    icon_fd.close()

    icon_svg = ICON_TRANSFORM + icon_payload + '\n</g>'

    tmp_path = os.path.join(user_activities_path, new_basename, 'activity',
                            'tmp.svg')
    tmp_icon_fd = open(tmp_path, 'w')
    tmp_icon_fd.write(XML_HEADER)
    tmp_icon_fd.write(SVG_START)
    tmp_icon_fd.write(icon_svg)
    tmp_icon_fd.write(badge_svg)
    tmp_icon_fd.write(SVG_END)
    tmp_icon_fd.close()

    os.remove(icon_path)
    os.rename(tmp_path, icon_path)


def _extract_svg_payload(fd):
    """Returns everything between <svg ...> and </svg>"""
    payload = ''
    looking_for_start_svg_token = True
    looking_for_close_token = True
    looking_for_end_svg_token = True
    for line in fd:
        if looking_for_start_svg_token:
            if line.find('<svg') < 0:
                continue
            looking_for_start_svg_token = False
            line = line.split('<svg', 1)[1]
        if looking_for_close_token:
            if line.find('>') < 0:
                continue
            looking_for_close_token = False
            line = line.split('>', 1)[1]
        if looking_for_end_svg_token:
            if line.find('</svg>') < 0:
                payload += line
                continue
            payload += line.split('</svg>')[0]
            break
    return payload



# Copyright (C) 2008 One Laptop Per Child
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import os
import logging
import subprocess
from gettext import gettext as _
import errno
import time

from gi.repository import Gio
from gi.repository import GLib

from jarabe import config
from jarabe.model.network import get_wireless_interfaces


_OFW_TREE = '/ofw'
_PROC_TREE = '/proc/device-tree'
_DMI_DIRECTORY = '/sys/class/dmi/id'

_logger = logging.getLogger('ControlPanel - AboutComputer')
_not_available = _('Not available')

_serial_no = None


def get_aboutcomputer():
    msg = 'Serial Number: %s \nBuild Number: %s \nFirmware Number: %s \n' \
        % (get_serial_number(), get_build_number(), get_firmware_number())
    return msg


def print_aboutcomputer():
    print(get_aboutcomputer())


def _get_serial_number():
    serial_no = _read_device_tree('serial-number')
    if serial_no is not None:
        return serial_no

    cmd = 'pkexec sugar-serial-number-helper'
    result, output, error, status = GLib.spawn_command_line_sync(cmd)
    if status != 0:
        return _not_available

    return output.decode().rstrip('\n')


def get_serial_number():
    global _serial_no

    if _serial_no is None:
        _serial_no = _get_serial_number()

    return _serial_no


def print_serial_number():
    serial_no = get_serial_number()
    if serial_no is None:
        serial_no = _not_available
    print(serial_no)


def get_build_number():
    build_no = _read_file('/boot/olpc_build')

    if build_no is None:
        build_no = _read_file('/etc/redhat-release')

    if build_no is None:
        try:
            popen = subprocess.Popen(['lsb_release', '-ds'],
                                     stdout=subprocess.PIPE)
        except OSError as e:
            if e.errno != errno.ENOENT:
                raise
        else:
            build_no, stderr_ = popen.communicate()
            build_no = build_no.strip().decode()

    if build_no is None or not build_no:
        build_no = _not_available

    return build_no


def print_build_number():
    print(get_build_number())


def get_firmware_number():
    firmware_no = _read_device_tree('openprom/model')
    if firmware_no is not None:
        # try to extract Open Firmware version from OLPC style version
        # string, e.g. "CL2   Q4B11  Q4B"
        if firmware_no.startswith('CL'):
            firmware_no = firmware_no[6:13].strip()
        ec_name = _read_device_tree('ec-name')
        if ec_name:
            ec_name = ec_name.replace('Ver:', '')
            firmware_no = '%(firmware)s with %(ec)s' % {
                'firmware': firmware_no, 'ec': ec_name}

    elif os.path.exists(os.path.join(_DMI_DIRECTORY, 'bios_version')):
        firmware_no = _read_file(os.path.join(_DMI_DIRECTORY, 'bios_version'))
    if firmware_no is None:
        firmware_no = _not_available
    return firmware_no


def get_hardware_model():
    settings = Gio.Settings.new('org.sugarlabs.extensions.aboutcomputer')
    model = settings.get_string('hardware-model')
    if not model:
        model = _read_device_tree('mfg-data/MN')
        sku = _read_device_tree('mfg-data/sk')
        if sku:
            if sku.startswith('SKU'):
                model = '%s, %s' % (model, sku)
            else:
                model = '%s, SKU%s' % (model, sku)
    return model


def get_secondary_licenses():
    licenses = []
    # Check if there are more licenses to display
    licenses_path = config.licenses_path
    if os.path.isdir(licenses_path):
        for file_name in os.listdir(licenses_path):
            try:
                file_path = os.path.join(licenses_path, file_name)
                with open(file_path) as f:
                    licenses.append(f.read())
            except IOError:
                logging.error('Error trying open %s', file_path)
    return licenses


def print_firmware_number():
    print(get_firmware_number())


def get_wireless_firmware():
    environment = os.environ.copy()
    environment['PATH'] = '%s:/usr/sbin' % (environment['PATH'], )
    firmware_info = {}

    wireless_interfaces = get_wireless_interfaces()
    if not wireless_interfaces:
        _logger.warning('NetworkManager knows of no wireless devices, '
                        'falling back to static list')
        wireless_interfaces = ['wlan0', 'eth0']

    for interface in get_wireless_interfaces():
        try:
            output = subprocess.Popen(['ethtool', '-i', interface],
                                      stdout=subprocess.PIPE,
                                      env=environment).stdout.readlines()
        except OSError:
            _logger.exception('Error running ethtool for %r', interface)
            continue

        try:
            for line in output:
                line = line.decode()
                if line.startswith('firmware'):
                    version = line.split()[1]
                if line.startswith('driver'):
                    driver = line.split()[1]
        except IndexError:
            _logger.exception('Error parsing ethtool output for %r',
                              interface)
            continue

        card = None
        if driver == 'mwifiex':
            card = 'mv8787, IEEE 802.11n 5GHz'
        if driver == 'libertas':
            if version.startswith('5.'):
                card = 'usb8388, IEEE 802.11g 2.4GHz'
            else:
                card = 'mv8686, IEEE 802.11g 2.4GHz'

        if card:
            firmware_info[interface] = '%s (%s, %s)' % (version, driver, card)
        else:
            firmware_info[interface] = '%s (%s)' % (version, driver)

    if not firmware_info:
        return _not_available

    if len(firmware_info) == 1:
        return list(firmware_info.values())[0]

    return ', '.join(['%(interface)s: %(info)s' %
                      {'interface': interface, 'info': info}
                      for interface, info in list(firmware_info.items())])


def print_wireless_firmware():
    print(get_wireless_firmware())


def _read_file(path):
    if os.access(path, os.R_OK) == 0:
        return None

    fd = open(path, 'r')
    value = fd.read()
    fd.close()
    if value:
        value = value.strip('\n')
        return value
    _logger.debug('No information in file or directory: %s', path)
    return None


def get_license():
    license_file = os.path.join(config.data_path, 'GPLv3')
    lang = os.environ['LANG']
    if lang.endswith('UTF-8'):
        lang = lang[:-6]

    try_file = license_file + '.' + lang
    if os.path.isfile(try_file):
        license_file = try_file
    else:
        try_file = license_file + '.' + lang.split('_')[0]
        if os.path.isfile(try_file):
            license_file = try_file

    try:
        fd = open(license_file)
        # remove 0x0c page breaks which can't be rendered in text views
        license_text = fd.read().replace('\x0c', '')
        fd.close()
    except IOError:
        license_text = _not_available
    return license_text


def _read_device_tree(path):
    value = _read_file(os.path.join(_PROC_TREE, path))
    if value:
        return value.strip('\x00')
    value = _read_file(os.path.join(_OFW_TREE, path))
    if value:
        return value.strip('\x00')
    return value


def days_from_last_update():

    last_update_seconds = -1
    # Get the number of seconds of the last update date.
    try:
        flag_file = '/var/lib/misc/last_os_update.stamp'
        if os.path.exists(flag_file):
            last_update_seconds = int(os.stat(flag_file).st_mtime)
    except IOError:
        _logger.error('couldn''t get last modification time')

    if last_update_seconds == -1:
        return -1

    now = time.time()
    days_from_last_update = (now - last_update_seconds) / (24 * 60 * 60)
    return int(days_from_last_update)



# Copyright (C) 2007, One Laptop Per Child
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
from urllib.parse import urlparse
import hashlib

from gi.repository import Gtk
from gi.repository import Gdk

from jarabe.frame.framewindow import FrameWindow
from jarabe.frame.clipboardtray import ClipboardTray

from jarabe.frame import clipboard


class ClipboardPanelWindow(FrameWindow):

    def __init__(self, frame, orientation):
        FrameWindow.__init__(self, orientation)

        self._frame = frame

        # Listening for new clipboard objects
        # NOTE: we need to keep a reference to Gtk.Clipboard in order to keep
        # listening to it.
        self._clipboard = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD)
        self._clipboard.connect('owner-change', self._owner_change_cb)

        self._clipboard_tray = ClipboardTray()
        self._clipboard_tray.show()
        self.append(self._clipboard_tray)

        # Receiving dnd drops
        self.drag_dest_set(0, [], 0)
        self.connect('drag-motion', self._clipboard_tray.drag_motion_cb)
        self.connect('drag-leave', self._clipboard_tray.drag_leave_cb)
        self.connect('drag-drop', self._clipboard_tray.drag_drop_cb)
        self.connect('drag-data-received',
                     self._clipboard_tray.drag_data_received_cb)

    def _owner_change_cb(self, x_clipboard, event):
        logging.debug('owner_change_cb')

        if self._clipboard_tray.owns_clipboard():
            return

        cb_service = clipboard.get_instance()

        result, targets = x_clipboard.wait_for_targets()
        cb_selections = []
        if not result:
            return

        target_is_uri = False
        for target in targets:
            if target not in ('TIMESTAMP', 'TARGETS',
                              'MULTIPLE', 'SAVE_TARGETS'):
                logging.debug('Asking for target %s.', target)
                if target == 'text/uri-list':
                    target_is_uri = True

                selection = x_clipboard.wait_for_contents(target)
                if not selection:
                    logging.warning('no data for selection target %s.', target)
                    continue
                cb_selections.append(selection)

        if target_is_uri:
            uri = selection.get_uris()[0]
            filename = uri[len('file://'):].strip()
            md5 = self._md5_for_file(filename)
            data_hash = hash(md5)
        else:
            data_hash = hash(selection.get_data())

        if len(cb_selections) > 0:
            key = cb_service.add_object(name="", data_hash=data_hash)
            if key is None:
                return
            cb_service.set_object_percent(key, percent=0)
            for selection in cb_selections:
                self._add_selection(key, selection)
            cb_service.set_object_percent(key, percent=100)

    def _md5_for_file(self, file_name):
        '''Calculate md5 for file data

        Calculating block wise to prevent issues with big files in memory
        '''
        block_size = 8192
        md5 = hashlib.md5()
        f = open(file_name, 'r')
        while True:
            data = f.read(block_size)
            if not data:
                break
            md5.update(data)
        f.close()
        return md5.digest()

    def _add_selection(self, key, selection):
        if not selection.get_data():
            logging.warning('no data for selection target %s.',
                            selection.get_data_type())
            return

        selection_type = str(selection.get_data_type())
        logging.debug('adding type ' + selection_type + '.')

        cb_service = clipboard.get_instance()
        if selection_type == 'text/uri-list':
            uris = selection.get_uris()

            if len(uris) > 1:
                raise NotImplementedError('Multiple uris in text/uri-list'
                                          ' still not supported.')
            uri = uris[0]
            scheme, netloc_, path_, parameters_, query_, fragment_ = \
                urlparse(uri)
            on_disk = (scheme == 'file')

            cb_service.add_object_format(key,
                                         selection_type,
                                         uri,
                                         on_disk)
        else:
            cb_service.add_object_format(key,
                                         selection_type,
                                         selection.get_data(),
                                         on_disk=False)



# Copyright (C) 2007, One Laptop Per Child
# Copyright (C) 2008-2013, Sugar Labs
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
from gettext import gettext as _
import time
import os

from gi.repository import GObject
from gi.repository import GLib
from gi.repository import Gtk
import json

from sugar3.graphics import style
from sugar3.graphics.xocolor import XoColor
from sugar3.graphics.icon import CanvasIcon, get_icon_file_name
from sugar3.graphics.icon import Icon, CellRendererIcon
from sugar3.graphics.alert import Alert, ConfirmationAlert
from sugar3.util import format_size
from sugar3.graphics.objectchooser import get_preview_pixbuf
from sugar3.activity.activity import PREVIEW_SIZE

from jarabe.journal.keepicon import KeepIcon
from jarabe.journal.palettes import ObjectPalette, BuddyPalette
from jarabe.journal import misc
from jarabe.journal import model
from jarabe.journal import journalwindow


class Separator(Gtk.VBox):

    def __init__(self, orientation):
        Gtk.VBox.__init__(
            self, background_color=style.COLOR_PANEL_GREY.get_gdk_color())


class BuddyList(Gtk.Alignment):

    def __init__(self, buddies):
        Gtk.Alignment.__init__(self)
        self.set(0, 0, 0, 0)

        hbox = Gtk.HBox()
        for buddy in buddies:
            nick_, color = buddy
            icon = CanvasIcon(icon_name='computer-xo',
                              xo_color=XoColor(color),
                              pixel_size=style.STANDARD_ICON_SIZE)
            icon.set_palette(BuddyPalette(buddy))
            hbox.pack_start(icon, True, True, 0)
        self.add(hbox)


class TextView(Gtk.TextView):

    def __init__(self):
        Gtk.TextView.__init__(self)
        text_buffer = Gtk.TextBuffer()
        self.set_buffer(text_buffer)
        self.set_left_margin(style.DEFAULT_PADDING)
        self.set_wrap_mode(Gtk.WrapMode.WORD_CHAR)


class CommentsView(Gtk.TreeView):
    __gsignals__ = {
        'comments-changed': (GObject.SignalFlags.RUN_FIRST, None, ([str])),
        'clicked': (GObject.SignalFlags.RUN_FIRST, None, [object]),
    }

    FROM = 'from'
    MESSAGE = 'message'
    ICON = 'icon'
    ICON_COLOR = 'icon-color'
    COMMENT_ICON = 0
    COMMENT_ICON_COLOR = 1
    COMMENT_FROM = 2
    COMMENT_MESSAGE = 3
    COMMENT_ERASE_ICON = 4
    COMMENT_ERASE_ICON_COLOR = 5

    def __init__(self):
        Gtk.TreeView.__init__(self)
        self.set_headers_visible(False)
        self._store = Gtk.ListStore(str, object, str, str, str, object)
        self._comments = []
        self._init_model()

    def update_comments(self, comments):
        self._store.clear()

        if comments:
            self._comments = json.loads(comments)
            for comment in self._comments:
                self._add_row(comment.get(self.FROM, ''),
                              comment.get(self.MESSAGE, ''),
                              comment.get(self.ICON, 'computer-xo'),
                              comment.get(self.ICON_COLOR, '#FFFFFF,#000000'))

    def _get_selected_row(self):
        selection = self.get_selection()
        return selection.get_selected()

    def _add_row(self, sender, message, icon_name, icon_color):
        self._store.append((get_icon_file_name(icon_name),
                            XoColor(icon_color),
                            sender,
                            message,
                            get_icon_file_name('list-remove'),
                            XoColor('#FFFFFF,#000000')))

    def _init_model(self):
        self.set_model(self._store)
        col = Gtk.TreeViewColumn()

        who_icon = CellRendererCommentIcon()
        col.pack_start(who_icon, False)
        col.add_attribute(who_icon, 'file-name', self.COMMENT_ICON)
        col.add_attribute(who_icon, 'xo-color', self.COMMENT_ICON_COLOR)

        who_text = Gtk.CellRendererText()
        col.pack_start(who_text, True)
        col.add_attribute(who_text, 'text', self.COMMENT_FROM)

        comment_text = Gtk.CellRendererText()
        col.pack_start(comment_text, True)
        col.add_attribute(comment_text, 'text', self.COMMENT_MESSAGE)

        erase_icon = CellRendererCommentIcon()
        erase_icon.connect('clicked', self._erase_comment_cb)
        col.pack_start(erase_icon, False)
        col.add_attribute(erase_icon, 'file-name', self.COMMENT_ERASE_ICON)
        col.add_attribute(
            erase_icon, 'xo-color', self.COMMENT_ERASE_ICON_COLOR)

        self.append_column(col)

    def _erase_comment_cb(self, widget, event):
        alert = Alert()

        entry = self.get_selection().get_selected()[1]
        erase_string = _('Erase')
        alert.props.title = erase_string
        alert.props.msg = _('Do you want to permanently erase \"%s\"?') \
            % self._store[entry][self.COMMENT_MESSAGE]

        icon = Icon(icon_name='dialog-cancel')
        alert.add_button(Gtk.ResponseType.CANCEL, _('Cancel'), icon)
        icon.show()

        ok_icon = Icon(icon_name='dialog-ok')
        alert.add_button(Gtk.ResponseType.OK, erase_string, ok_icon)
        ok_icon.show()

        alert.connect('response', self._erase_alert_response_cb, entry)

        journalwindow.get_journal_window().add_alert(alert)
        alert.show()

    def _erase_alert_response_cb(self, alert, response_id, entry):
        journalwindow.get_journal_window().remove_alert(alert)

        if response_id is Gtk.ResponseType.OK:
            self._store.remove(entry)

            # Regenerate comments from current contents of store
            self._comments = []
            for entry in self._store:
                self._comments.append({
                    self.FROM: entry[self.COMMENT_FROM],
                    self.MESSAGE: entry[self.COMMENT_MESSAGE],
                    self.ICON: entry[self.COMMENT_ICON],
                    self.ICON_COLOR: '[%s]' % (
                        entry[self.COMMENT_ICON_COLOR].to_string()),
                })

            self.emit('comments-changed', json.dumps(self._comments))


class CellRendererCommentIcon(CellRendererIcon):

    def __init__(self):
        CellRendererIcon.__init__(self)

        self.props.width = style.SMALL_ICON_SIZE
        self.props.height = style.SMALL_ICON_SIZE
        self.props.size = style.SMALL_ICON_SIZE
        self.props.stroke_color = style.COLOR_BUTTON_GREY.get_svg()
        self.props.fill_color = style.COLOR_BLACK.get_svg()
        self.props.mode = Gtk.CellRendererMode.ACTIVATABLE


class BaseExpandedEntry(GObject.GObject):

    def __init__(self):
        # Create a header
        self._keep_icon = None
        self._keep_sid = None
        self._icon = None
        self._icon_box = None
        self._title = None
        self._date = None

    def create_header(self):
        header = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)

        self._keep_icon = self._create_keep_icon()
        header.pack_start(self._keep_icon, False, False, style.DEFAULT_SPACING)

        self._icon_box = Gtk.HBox()
        header.pack_start(self._icon_box, False, False, style.DEFAULT_SPACING)

        self._title = self._create_title()
        header.pack_start(self._title, True, True, 0)

        # TODO: create a version list popup instead of a date label
        self._date = self._create_date()
        header.pack_start(self._date, False, False, style.DEFAULT_SPACING)

        return header

    def _create_keep_icon(self):
        keep_icon = KeepIcon()
        return keep_icon

    def _create_title(self):
        entry = Gtk.Entry()
        return entry

    def _create_date(self):
        date = Gtk.Label()
        return date


class ExpandedEntry(Gtk.EventBox, BaseExpandedEntry):

    def __init__(self, journalactivity):
        BaseExpandedEntry.__init__(self)
        self._journalactivity = journalactivity
        Gtk.EventBox.__init__(self)
        self._vbox = Gtk.VBox()
        self.add(self._vbox)

        self.in_focus = False
        self._metadata = None
        self._update_title_sid = None

        self.modify_bg(Gtk.StateType.NORMAL, style.COLOR_WHITE.get_gdk_color())

        self._header = self.create_header()
        self._vbox.pack_start(self._header, False, False,
                              style.DEFAULT_SPACING * 2)
        self._keep_sid = self._keep_icon.connect(
            'toggled', self._keep_icon_toggled_cb)
        self._title.connect('activate', self._title_entered)
        self._title.connect(
            'focus-out-event', self._focus_out_cb)
        self._title.connect(
            'focus-in-event', self._focus_in_cb)

        if Gtk.Widget.get_default_direction() == Gtk.TextDirection.RTL:
            # Reverse header children.
            for child in self._header.get_children():
                self._header.reorder_child(child, 0)

        # Create a two-column body
        body_box = Gtk.EventBox()
        body_box.set_border_width(style.DEFAULT_SPACING)
        body_box.modify_bg(Gtk.StateType.NORMAL,
                           style.COLOR_WHITE.get_gdk_color())
        self._vbox.pack_start(body_box, True, True, 0)
        body = Gtk.HBox()
        body_box.add(body)

        first_column = Gtk.VBox()
        body.pack_start(first_column, False, False, style.DEFAULT_SPACING)

        second_column = Gtk.VBox()
        body.pack_start(second_column, True, True, 0)

        # First body column
        self._preview_box = Gtk.Frame()
        style_context = self._preview_box.get_style_context()
        style_context.add_class('journal-preview-box')
        first_column.pack_start(self._preview_box, False, True, 0)

        self._technical_box = Gtk.VBox()
        first_column.pack_start(self._technical_box, False, False, 0)

        # Second body column
        description_box, self._description = self._create_description()
        second_column.pack_start(description_box, True, True,
                                 style.DEFAULT_SPACING)

        tags_box, self._tags = self._create_tags()
        second_column.pack_start(tags_box, True, True,
                                 style.DEFAULT_SPACING)

        comments_box, self._comments = self._create_comments()
        second_column.pack_start(comments_box, True, True,
                                 style.DEFAULT_SPACING)

        self._buddy_list = Gtk.VBox()
        second_column.pack_start(self._buddy_list, True, False, 0)
        self.show_all()

    def set_metadata(self, metadata):
        if self._metadata == metadata:
            return
        self._metadata = metadata

        self._keep_icon.handler_block(self._keep_sid)
        self._keep_icon.set_active(int(metadata.get('keep', 0)) == 1)
        self._keep_icon.handler_unblock(self._keep_sid)

        self._icon = self._create_icon()
        for child in self._icon_box.get_children():
            self._icon_box.remove(child)
            # FIXME: self._icon_box.foreach(self._icon_box.remove)
        self._icon_box.pack_start(self._icon, False, False, 0)

        self._date.set_text(misc.get_date(metadata))

        self._title.set_text(metadata.get('title', _('Untitled')))

        if self._preview_box.get_child():
            self._preview_box.remove(self._preview_box.get_child())
        self._preview_box.add(self._create_preview())

        for child in self._technical_box.get_children():
            self._technical_box.remove(child)
            # FIXME: self._technical_box.foreach(self._technical_box.remove)
        self._technical_box.pack_start(self._create_technical(),
                                       False, False, style.DEFAULT_SPACING)

        for child in self._buddy_list.get_children():
            self._buddy_list.remove(child)
            # FIXME: self._buddy_list.foreach(self._buddy_list.remove)
        self._buddy_list.pack_start(self._create_buddy_list(), False, False,
                                    style.DEFAULT_SPACING)

        description = metadata.get('description', '')
        self._description.get_buffer().set_text(description)
        tags = metadata.get('tags', '')
        self._tags.get_buffer().set_text(tags)
        comments = metadata.get('comments', '')
        self._comments.update_comments(comments)

    def _create_icon(self):
        icon = CanvasIcon(file_name=misc.get_icon_name(self._metadata))
        icon.connect_after('activate', self.__icon_activate_cb)

        if misc.is_activity_bundle(self._metadata):
            xo_color = XoColor('%s,%s' % (style.COLOR_BUTTON_GREY.get_svg(),
                                          style.COLOR_TRANSPARENT.get_svg()))
        else:
            xo_color = misc.get_icon_color(self._metadata)
        icon.props.xo_color = xo_color

        icon.set_palette(ObjectPalette(self._journalactivity, self._metadata))

        return icon

    def _create_preview(self):

        box = Gtk.EventBox()
        box.modify_bg(Gtk.StateType.NORMAL, style.COLOR_WHITE.get_gdk_color())

        metadata = self._metadata
        pixbuf = get_preview_pixbuf(metadata.get('preview', ''))
        has_preview = pixbuf is not None

        if has_preview:
            im = Gtk.Image()
            im.set_from_pixbuf(pixbuf)
            box.add(im)
            im.show()
        else:
            label = Gtk.Label()
            label.set_text(_('No preview'))
            width, height = PREVIEW_SIZE[0], PREVIEW_SIZE[1]
            label.set_size_request(width, height)
            box.add(label)
            label.show()

        box.connect_after('button-release-event',
                          self._preview_box_button_release_event_cb)
        return box

    def _create_technical(self):
        vbox = Gtk.VBox()
        vbox.props.spacing = style.DEFAULT_SPACING

        if 'filesize' in self._metadata:
            filesize = self._metadata['filesize']
        else:
            filesize = model.get_file_size(self._metadata['uid'])

        lines = [
            _('Kind: %s') % (self._metadata.get('mime_type') or _('Unknown'),),
            _('Date: %s') % (self._format_date(),),
            _('Size: %s') % (format_size(int(filesize)))
        ]

        for line in lines:
            linebox = Gtk.HBox()
            vbox.pack_start(linebox, False, False, 0)

            text = Gtk.Label()
            text.set_markup('<span foreground="%s">%s</span>' % (
                style.COLOR_BUTTON_GREY.get_html(), line))
            linebox.pack_start(text, False, False, 0)

        return vbox

    def _format_date(self):
        if 'timestamp' in self._metadata:
            try:
                timestamp = float(self._metadata['timestamp'])
            except (ValueError, TypeError):
                logging.warning('Invalid timestamp for %r: %r',
                                self._metadata['uid'],
                                self._metadata['timestamp'])
            else:
                return time.strftime('%x', time.localtime(timestamp))
        return _('No date')

    def _create_buddy_list(self):

        vbox = Gtk.VBox()
        vbox.props.spacing = style.DEFAULT_SPACING

        text = Gtk.Label()
        text.set_markup('<span foreground="%s">%s</span>' % (
            style.COLOR_BUTTON_GREY.get_html(), _('Participants:')))
        halign = Gtk.Alignment.new(0, 0, 0, 0)
        halign.add(text)
        vbox.pack_start(halign, False, False, 0)

        if self._metadata.get('buddies'):
            buddies = list(json.loads(self._metadata['buddies']).values())
            vbox.pack_start(BuddyList(buddies), False, False, 0)
            return vbox
        return vbox

    def _create_scrollable(self, widget, label=None):
        vbox = Gtk.VBox()
        vbox.props.spacing = style.DEFAULT_SPACING

        if label is not None:
            text = Gtk.Label()
            text.set_markup('<span foreground="%s">%s</span>' % (
                style.COLOR_BUTTON_GREY.get_html(), label))

            halign = Gtk.Alignment.new(0, 0, 0, 0)
            halign.add(text)
            vbox.pack_start(halign, False, False, 0)

        scrolled_window = Gtk.ScrolledWindow()
        scrolled_window.set_policy(Gtk.PolicyType.AUTOMATIC,
                                   Gtk.PolicyType.AUTOMATIC)
        scrolled_window.set_shadow_type(Gtk.ShadowType.ETCHED_IN)
        scrolled_window.add(widget)
        vbox.pack_start(scrolled_window, True, True, 0)

        return vbox

    def _create_description(self):
        widget = TextView()
        widget.connect('focus-in-event', self._focus_in_cb)
        widget.connect('focus-out-event',
                       self._description_tags_focus_out_event_cb)
        return self._create_scrollable(widget, label=_('Description:')), widget

    def _create_tags(self):
        widget = TextView()
        widget.connect('focus-in-event', self._focus_in_cb)
        widget.connect('focus-out-event',
                       self._description_tags_focus_out_event_cb)
        return self._create_scrollable(widget, label=_('Tags:')), widget

    def _create_comments(self):
        widget = CommentsView()
        widget.connect('comments-changed', self._comments_changed_cb)
        widget.connect('focus-in-event', self._focus_in_cb)
        widget.connect('focus-out-event', self._focus_out_cb)
        return self._create_scrollable(widget, label=_('Comments:')), widget

    def _focus_in_cb(self, widget, event):
        self.in_focus = True

    def _focus_out_cb(self, widget, event):
        self.in_focus = False

    def _title_entered(self, widget):
        self._update_entry()
        self._title.hide()
        self._title.show()

    def _title_notify_text_cb(self, entry, pspec):
        if not self._update_title_sid:
            self._update_title_sid = \
                GLib.timeout_add_seconds(1,
                                         self._update_title_cb)

    def _title_changed_event_cb(self, widget):
        old_title = self._metadata.get('title', None)
        new_title = self._title.get_text()
        if old_title != new_title:
            if new_title == '' or new_title.isspace():
                alert = ConfirmationAlert()
                alert.props.title = _('Empty title')
                alert.props.msg = _('The title is usually not left empty')
                alert.connect(
                    'response',
                    self._title_alert_response_cb,
                    old_title,
                    self._metadata.get('title_set_by_user', 0)
                )
                journalwindow.get_journal_window().add_alert(alert)
                alert.show()

            self._update_entry()

    def _title_alert_response_cb(self, alert, response_id, old_title, old_title_set_by_user):
        journalwindow.get_journal_window().remove_alert(alert)

        if response_id is Gtk.ResponseType.CANCEL:
            self._title.set_text(old_title)
            self._icon.palette.props.primary_text = old_title
            self._metadata['title'] = old_title
            self._metadata['title_set_by_user'] = old_title_set_by_user
            self._update_entry(needs_update=True)

    def _description_tags_focus_out_event_cb(self, text_view, event):
        self._update_entry()

    def _comments_changed_cb(self, event, comments):
        self._metadata['comments'] = comments
        self._write_entry()

    def _update_entry(self, needs_update=False):
        self.in_focus = False
        if not model.is_editable(self._metadata):
            return

        old_title = self._metadata.get('title', None)
        new_title = self._title.get_text()
        if old_title != new_title:
            self._icon.palette.props.primary_text = new_title
            self._metadata['title'] = new_title
            self._metadata['title_set_by_user'] = '1'
            needs_update = True


        bounds = self._tags.get_buffer().get_bounds()
        old_tags = self._metadata.get('tags', None)
        new_tags = self._tags.get_buffer().get_text(bounds[0], bounds[1],
                                                    include_hidden_chars=False)

        if old_tags != new_tags:
            self._metadata['tags'] = new_tags
            needs_update = True

        bounds = self._description.get_buffer().get_bounds()
        old_description = self._metadata.get('description', None)
        new_description = self._description.get_buffer().get_text(
            bounds[0], bounds[1], include_hidden_chars=False)
        if old_description != new_description:
            self._metadata['description'] = new_description
            needs_update = True

        if needs_update:
            self._write_entry()

        self._update_title_sid = None

    def _write_entry(self):
        if self._metadata.get('mountpoint', '/') == '/':
            model.write(self._metadata, update_mtime=False)
        else:
            old_file_path = os.path.join(
                self._metadata['mountpoint'],
                model.get_file_name(self._metadata['title'],
                                    self._metadata['mime_type']))
            model.write(self._metadata, file_path=old_file_path,
                        update_mtime=False)

    def _keep_icon_toggled_cb(self, keep_icon):
        if keep_icon.get_active():
            self._metadata['keep'] = '1'
        else:
            self._metadata['keep'] = '0'
        self._update_entry(needs_update=True)

    def __icon_activate_cb(self, button):
        misc.resume(self._metadata,
                    alert_window=journalwindow.get_journal_window())
        return True

    def _preview_box_button_release_event_cb(self, button, event):
        logging.debug('_preview_box_button_release_event_cb')
        misc.resume(self._metadata,
                    alert_window=journalwindow.get_journal_window())
        return True



# Copyright (C) 2008 One Laptop Per Child
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import math

from gi.repository import GObject
from gi.repository import GLib

from sugar3.graphics.icon import Icon
from sugar3.graphics import style
from sugar3.graphics.icon import CanvasIcon


_INTERVAL = 100
_STEP = math.pi / 10  # must be a fraction of pi, for clean caching
_MINIMAL_ALPHA_VALUE = 0.33


class Pulser(object):

    def __init__(self, icon, interval=_INTERVAL):
        self._pulse_hid = None
        self._icon = icon
        self._interval = interval
        self._phase = 0
        self._start_scale = 1.0
        self._end_scale = 1.0
        self._zoom_steps = 1
        self._current_zoom_step = 1
        self._current_scale_step = 1

    def set_zooming(self, start_scale, end_scale, zoom_steps):
        """ Set start and end scale and number of steps in zoom animation """
        self._start_scale = start_scale
        self._end_scale = end_scale
        self._zoom_steps = zoom_steps
        self._current_scale_step = abs(self._start_scale - self._end_scale) / \
            self._zoom_steps
        self._icon.scale = self._start_scale

    def start(self, restart=False):
        if restart:
            self._phase = 0
        if self._pulse_hid is None:
            self._pulse_hid = GLib.timeout_add(self._interval,
                                               self.__pulse_cb)
        if self._start_scale != self._end_scale:
            self._icon.scale = self._start_scale + \
                self._current_scale_step * self._current_zoom_step

    def stop(self):
        if self._pulse_hid is not None:
            GLib.source_remove(self._pulse_hid)
            self._pulse_hid = None
        self._icon.xo_color = self._icon.get_base_color()
        self._phase = 0
        self._icon.alpha = 1.0

    def update(self):
        self._icon.xo_color = self._icon.base_color
        self._icon.alpha = _MINIMAL_ALPHA_VALUE + \
            (1 - _MINIMAL_ALPHA_VALUE) * (math.cos(self._phase) + 1) / 2

    def __pulse_cb(self):
        self._phase += _STEP
        if self._current_zoom_step <= self._zoom_steps and \
                self._start_scale != self._end_scale:
            self._icon.scale = self._start_scale + \
                self._current_scale_step * self._current_zoom_step
            self._current_zoom_step += 1
        self.update()
        return True


class PulsingIcon(Icon):
    __gtype_name__ = 'SugarPulsingIcon'

    def __init__(self, interval=_INTERVAL, **kwargs):
        self._pulser = Pulser(self, interval)
        self._base_color = None
        self._pulse_color = None
        self._paused = False
        self._pulsing = False

        Icon.__init__(self, **kwargs)

        self._palette = None
        self.connect('destroy', self.__destroy_cb)

    def set_pulse_color(self, pulse_color):
        self._pulse_color = pulse_color
        self._pulser.update()

    def get_pulse_color(self):
        return self._pulse_color

    pulse_color = GObject.Property(
        type=object, getter=get_pulse_color, setter=set_pulse_color)

    def set_base_color(self, base_color):
        self._base_color = base_color
        self._pulser.update()

    def get_base_color(self):
        return self._base_color

    def set_zooming(self, start_size=style.SMALL_ICON_SIZE,
                    end_size=style.XLARGE_ICON_SIZE,
                    zoom_steps=10):
        if start_size > end_size:
            start_scale = 1.0
            end_scale = float(end_size) / start_size
        else:
            start_scale = float(start_size) / end_size
            end_scale = 1.0
        self._pulser.set_zooming(start_scale, end_scale, zoom_steps)

    base_color = GObject.Property(
        type=object, getter=get_base_color, setter=set_base_color)

    def set_paused(self, paused):
        self._paused = paused

        if self._paused:
            self._pulser.stop()
        else:
            self._pulser.start(restart=False)

    def get_paused(self):
        return self._paused

    paused = GObject.Property(
        type=bool, default=False, getter=get_paused, setter=set_paused)

    def set_pulsing(self, pulsing):
        self._pulsing = pulsing

        if self._pulsing:
            self._pulser.start(restart=True)
        else:
            self._pulser.stop()

    def get_pulsing(self):
        return self._pulsing

    pulsing = GObject.Property(
        type=bool, default=False, getter=get_pulsing, setter=set_pulsing)

    def _get_palette(self):
        return self._palette

    def _set_palette(self, palette):
        if self._palette is not None:
            self._palette.props.invoker = None
        self._palette = palette

    palette = property(_get_palette, _set_palette)

    def __destroy_cb(self, icon):
        self._pulser.stop()
        if self._palette is not None:
            self._palette.destroy()


class EventPulsingIcon(CanvasIcon):
    __gtype_name__ = 'SugarEventPulsingIcon'

    def __init__(self, interval=_INTERVAL, **kwargs):
        self._pulser = Pulser(self, interval)
        self._base_color = None
        self._pulse_color = None
        self._paused = False
        self._pulsing = False

        CanvasIcon.__init__(self, **kwargs)

        self.connect('destroy', self.__destroy_cb)

    def __destroy_cb(self, box):
        self._pulser.stop()

    def set_pulse_color(self, pulse_color):
        self._pulse_color = pulse_color
        self._pulser.update()

    def get_pulse_color(self):
        return self._pulse_color

    pulse_color = GObject.Property(
        type=object, getter=get_pulse_color, setter=set_pulse_color)

    def set_base_color(self, base_color):
        self._base_color = base_color
        self._pulser.update()

    def get_base_color(self):
        return self._base_color

    base_color = GObject.Property(
        type=object, getter=get_base_color, setter=set_base_color)

    def set_paused(self, paused):
        self._paused = paused

        if self._paused:
            self._pulser.stop()
        elif self._pulsing:
            self._pulser.start(restart=False)

    def get_paused(self):
        return self._paused

    paused = GObject.Property(
        type=bool, default=False, getter=get_paused, setter=set_paused)

    def set_pulsing(self, pulsing):
        self._pulsing = pulsing
        if self._paused:
            return

        if self._pulsing:
            self._pulser.start(restart=True)
        else:
            self._pulser.stop()

    def get_pulsing(self):
        return self._pulsing

    pulsing = GObject.Property(
        type=bool, default=False, getter=get_pulsing, setter=set_pulsing)



# Copyright (C) 2006-2007 Red Hat, Inc.
# Copyright (C) 2010 Collabora Ltd. <http://www.collabora.co.uk/>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""D-bus service providing access to the shell's functionality"""

import dbus
from gi.repository import Gtk

from jarabe.model import shell
from jarabe.model import bundleregistry


_DBUS_SERVICE = 'org.laptop.Shell'
_DBUS_SHELL_IFACE = 'org.laptop.Shell'
_DBUS_PATH = '/org/laptop/Shell'


class UIService(dbus.service.Object):
    """Provides d-bus service to script the shell's operations

    Uses a shell_model object to observe events such as changes to:

        * nickname
        * colour
        * icon
        * currently active activity

    and pass the event off to the methods in the dbus signature.

    Key method here at the moment is add_bundle, which is used to
    do a run-time registration of a bundle using it's application path.

    XXX At the moment the d-bus service methods do not appear to do
    anything other than add_bundle
    """

    def __init__(self):
        bus = dbus.SessionBus()
        bus_name = dbus.service.BusName(_DBUS_SERVICE, bus=bus)
        dbus.service.Object.__init__(self, bus_name, _DBUS_PATH)

        self._shell_model = shell.get_model()

    @dbus.service.method(_DBUS_SHELL_IFACE,
                         in_signature='s', out_signature='s')
    def GetBundlePath(self, bundle_id):
        bundle = bundleregistry.get_registry().get_bundle(bundle_id)
        if bundle:
            return bundle.get_path()
        return ''

    @dbus.service.method(_DBUS_SHELL_IFACE,
                         in_signature='s', out_signature='b')
    def ActivateActivity(self, activity_id):
        """Switch to the window related to this activity_id and return a
        boolean indicating if there is a real (ie. not a launcher window)
        activity already open.
        """
        activity = self._shell_model.get_activity_by_id(activity_id)

        if activity is not None and activity.get_window() is not None:
            activity.get_window().activate(Gtk.get_current_event_time())
            return self._shell_model.get_launcher(activity_id) is None

        return False

    @dbus.service.method(_DBUS_SHELL_IFACE,
                         in_signature='ss', out_signature='')
    def NotifyLaunch(self, bundle_id, activity_id):
        shell.get_model().notify_launch(activity_id, bundle_id)

    @dbus.service.method(_DBUS_SHELL_IFACE,
                         in_signature='s', out_signature='')
    def NotifyLaunchFailure(self, activity_id):
        shell.get_model().notify_launch_failed(activity_id)



# Copyright (C) 2016, Abhijit Patel
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from gettext import gettext as _

from gi.repository import GObject
from gi.repository import Gtk
from gi.repository import Gio

from jarabe.journal.expandedentry import TextView
from jarabe.journal.expandedentry import BaseExpandedEntry
from jarabe.journal.detailview import BackBar
from jarabe.journal.listview import ListView
from jarabe.journal import model

from sugar3.graphics.xocolor import XoColor
from sugar3.graphics import style
from sugar3.graphics.icon import Icon

_SERVICE_NAME = 'org.laptop.Activity'
_SERVICE_PATH = '/org/laptop/Activity'
_SERVICE_INTERFACE = 'org.laptop.Activity'


class ProjectView(Gtk.EventBox, BaseExpandedEntry):

    __gsignals__ = {
        'go-back-clicked': (GObject.SignalFlags.RUN_FIRST, None, ([])),
    }

    def __init__(self, **kwargs):
        Gtk.EventBox.__init__(self)
        BaseExpandedEntry.__init__(self)
        self.project_metadata = None
        self._service = None
        self._activity_id = None
        self._project = None
        self.modify_bg(Gtk.StateType.NORMAL, style.COLOR_WHITE.get_gdk_color())

        self._vbox = Gtk.VBox()
        self.add(self._vbox)

        back_bar = BackBar()
        back_bar.connect('button-release-event',
                         self.__back_bar_release_event_cb)
        self._vbox.pack_start(back_bar, False, True, 0)

        header = self.create_header()
        self._vbox.pack_start(header, False, False, style.DEFAULT_SPACING * 2)
        header.show()

        description_box, self._description = self._create_description()
        self._vbox.pack_start(description_box, False, True,
                              style.DEFAULT_SPACING / 3)

        self._title.connect('focus-out-event', self._title_focus_out_event_cb)

        settings = Gio.Settings.new('org.sugarlabs.user')
        icon_color = settings.get_string('color')

        self._icon = Icon(icon_name='project-box',
                          pixel_size=style.MEDIUM_ICON_SIZE)
        self._icon.xo_color = XoColor(icon_color)
        self._icon_box.pack_start(self._icon, False, False, 0)

    def get_vbox(self):
        return self._vbox

    def create_list_view_project(self):
        self._list_view_project = ListView(self)
        return self._list_view_project

    def get_list_view(self):
        return self._list_view_project

    def get_mount_point(self):
        return '/'

    def __back_bar_release_event_cb(self, back_bar, event):
        self.emit('go-back-clicked')
        return False

    def set_project_metadata(self, project_metadata):
        self.project_metadata = project_metadata

        description = project_metadata.get('description', '')
        self._description.get_buffer().set_text(description)
        self._title.set_text(project_metadata.get('title', ''))

    def _title_focus_out_event_cb(self, entry, event):
        self._update_entry()

    def _create_description(self):
        widget = TextView()
        widget.connect('focus-out-event',
                       self._description_tags_focus_out_event_cb)
        return self._create_scrollable(widget, label=_('Description:')), widget

    def _description_tags_focus_out_event_cb(self, text_view, event):
        self._update_entry()

    def _update_entry(self):
        # updating description
        bounds = self._description.get_buffer().get_bounds()
        old_description = self.project_metadata.get('description', None)
        new_description = self._description.get_buffer().get_text(
            bounds[0], bounds[1], include_hidden_chars=False)

        if old_description != new_description:
            self.project_metadata['description'] = new_description
            model.write(self.project_metadata)

        new_title = self._title.get_text()
        old_title = self.project_metadata.get('title', '')

        if old_title != new_title:
            self.project_metadata['title'] = new_title
            model.write(self.project_metadata)

    def _create_scrollable(self, widget, label=None):
        vbox = Gtk.VBox()
        vbox.props.spacing = style.DEFAULT_SPACING

        if label is not None:
            text = Gtk.Label()
            text.set_markup('<span foreground="%s">%s</span>' % (
                style.COLOR_BUTTON_GREY.get_html(), label))

            halign = Gtk.Alignment.new(0, 0, 0, 0)
            halign.add(text)
            vbox.pack_start(halign, False, False, 0)

        scrolled_window = Gtk.ScrolledWindow()
        scrolled_window.set_policy(Gtk.PolicyType.AUTOMATIC,
                                   Gtk.PolicyType.AUTOMATIC)
        scrolled_window.set_shadow_type(Gtk.ShadowType.ETCHED_IN)
        scrolled_window.add(widget)
        vbox.pack_start(scrolled_window, True, True, 0)

        return vbox
